"""If the title is empty, try to extract it from the filename
(possibly also extract track and artist)
"""

import os
import re

from beets import plugins
from beets.util import displayable_path

# Filename field extraction patterns.
PATTERNS = [
    # Useful patterns.
    (
        r"^(?P<track>\d+)\.?\s*-\s*(?P<artist>.+?)\s*-\s*(?P<title>.+?)"
        r"(\s*-\s*(?P<tag>.*))?$"
    ),
    r"^(?P<artist>.+?)\s*-\s*(?P<title>.+?)(\s*-\s*(?P<tag>.*))?$",
    r"^(?P<track>\d+)\.?[\s_-]+(?P<title>.+)$",
    r"^(?P<title>.+) by (?P<artist>.+)$",
    r"^(track ?)?(?P<track>\d+).*$",
    r"^(?P<title>.+)$",
]

# Titles considered "empty" and in need of replacement.
BAD_TITLE_PATTERNS = [r"^$"]


def equal(seq):
    """Determine whether a sequence holds identical elements."""
    return len(set(seq)) <= 1


def equal_fields(matchdict, field):
    """Do all items in `matchdict`, whose values are dictionaries, have
    the same value for `field`? (If they do, the field is probably not
    the title.)
    """
    return equal(m[field] for m in matchdict.values())


def all_matches(names, pattern):
    """If all the filenames in the item/filename mapping match the
    pattern, return a dictionary mapping the items to dictionaries
    giving the value for each named subpattern in the match. Otherwise,
    return None.
    """
    matches = {}
    for item, name in names.items():
        m = re.match(pattern, name, re.IGNORECASE)
        if m and m.groupdict():
            # Only yield a match when the regex applies *and* has
            # capture groups. Otherwise, no information can be extracted
            # from the filename.
            matches[item] = m.groupdict()
        else:
            return None
    return matches


def bad_title(title):
    """Determine whether a given title is "bad" (empty or otherwise
    meaningless) and in need of replacement.
    """
    for pat in BAD_TITLE_PATTERNS:
        if re.match(pat, title, re.IGNORECASE):
            return True
    return False


def apply_matches(d, log):
    """Given a mapping from items to field dicts, apply the fields to
    the objects.
    """
    some_map = next(iter(d.values()))
    keys = some_map.keys()

    # Only proceed if the "tag" field is equal across all filenames.
    if "tag" in keys and not equal_fields(d, "tag"):
        return

    # Given both an "artist" and "title" field, assume that one is
    # *actually* the artist, which must be uniform, and use the other
    # for the title. This, of course, won't work for VA albums.
    # Only check for "artist": patterns containing it, also contain "title"
    if "artist" in keys:
        if equal_fields(d, "artist"):
            artist = some_map["artist"]
            title_field = "title"
        elif equal_fields(d, "title"):
            artist = some_map["title"]
            title_field = "artist"
        else:
            # Both vary. Abort.
            return

        for item in d:
            if not item.artist:
                item.artist = artist
                log.info("Artist replaced with: {.artist}", item)
    # otherwise, if the pattern contains "title", use that for title_field
    elif "title" in keys:
        title_field = "title"
    else:
        title_field = None

    # Apply the title and track, if any.
    for item in d:
        if title_field and bad_title(item.title):
            item.title = str(d[item][title_field])
            log.info("Title replaced with: {.title}", item)

        if "track" in d[item] and item.track == 0:
            item.track = int(d[item]["track"])
            log.info("Track replaced with: {.track}", item)


# Plugin structure and hook into import process.


class FromFilenamePlugin(plugins.BeetsPlugin):
    def __init__(self):
        super().__init__()
        self.register_listener("import_task_start", self.filename_task)

    def filename_task(self, task, session):
        """Examine each item in the task to see if we can extract a title
        from the filename. Try to match all filenames to a number of
        regexps, starting with the most complex patterns and successively
        trying less complex patterns. As soon as all filenames match the
        same regex we can make an educated guess of which part of the
        regex that contains the title.
        """
        items = task.items if task.is_album else [task.item]

        # Look for suspicious (empty or meaningless) titles.
        missing_titles = sum(bad_title(i.title) for i in items)

        if missing_titles:
            # Get the base filenames (no path or extension).
            names = {}
            for item in items:
                path = displayable_path(item.path)
                name, _ = os.path.splitext(os.path.basename(path))
                names[item] = name

            # Look for useful information in the filenames.
            for pattern in PATTERNS:
                self._log.debug(f"Trying pattern: {pattern}")
                d = all_matches(names, pattern)
                if d:
                    apply_matches(d, self._log)



"""This module includes various helpers that provide fixtures, capture
information or mock the environment.

- `has_program` checks the presence of a command on the system.

- The `ImportSessionFixture` allows one to run importer code while
  controlling the interactions through code.

- The `TestHelper` class encapsulates various fixtures that can be set up.
"""

from __future__ import annotations

import importlib.util
import os
import os.path
import shutil
import subprocess
import sys
import unittest
from contextlib import contextmanager
from copy import deepcopy
from dataclasses import dataclass
from functools import cache, cached_property
from pathlib import Path
from tempfile import gettempdir, mkdtemp, mkstemp
from typing import TYPE_CHECKING, Any, ClassVar, Literal
from unittest.mock import Mock, patch

import pytest
from mediafile import Image, MediaFile

import beets
import beets.plugins
from beets import importer, util
from beets.autotag import AlbumInfo, TrackInfo
from beets.importer import ImportSession
from beets.library import Item, Library
from beets.test import _common
from beets.ui.commands.import_.session import TerminalImportSession
from beets.util import (
    MoveOperation,
    bytestring_path,
    clean_module_tempdir,
    syspath,
)
from beets.util.functemplate import template

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Sequence
    from types import TracebackType
    from unittest.mock import _patch

    from confuse import ConfigSource
    from requests_mock.mocker import Mocker
    from typing_extensions import Self

    from beets.autotag import AlbumMatch, TrackMatch
    from beets.library import Album

RUNNING_IN_CI = os.environ.get("GITHUB_ACTIONS") == "true"


def has_program(cmd: str, args: Iterable[str] = ("--version",)) -> bool:
    """Returns `True` if `cmd` can be executed."""
    full_cmd = [cmd, *args]
    try:
        with open(os.devnull, "wb") as devnull:
            subprocess.check_call(
                full_cmd, stderr=devnull, stdout=devnull, stdin=devnull
            )
    except OSError:
        return False
    except subprocess.CalledProcessError:
        return False
    else:
        return True


@cache
def is_importable(modname: str) -> bool:
    return bool(importlib.util.find_spec(modname))


def check_reflink_support(path: str) -> bool:
    try:
        import reflink
    except ImportError:
        return False

    return reflink.supported_at(path)


NEEDS_REFLINK = pytest.mark.skipif(
    not check_reflink_support(gettempdir()), reason="need reflink"
)
NEEDS_FFPROBE = pytest.mark.skipif(
    not has_program("ffprobe", ("-version",)) and not RUNNING_IN_CI,
    reason="ffprobe (ffmpeg) is not available",
)


class ConfigMixin:
    """Provide isolated configuration for tests."""

    _default_config_sources: ClassVar[list[ConfigSource] | None] = None

    @classmethod
    def default_config_sources(cls) -> list[ConfigSource]:
        """Return a reusable default configuration baseline.

        This way, we only need to call very expensive ``config.read`` once per
        test session.

        NOTE: we're not using ``util.cached_classproperty`` here because its cache is
        reset on every test.
        """
        if cls._default_config_sources is not None:
            return deepcopy(cls._default_config_sources)

        config = beets.IncludeLazyConfig("beets", beets.__name__)
        config.read(user=False, defaults=True)

        config["plugins"] = []
        config["verbose"] = 2
        config["ui"]["color"] = False
        config["threaded"] = False
        config["create_backup_before_migrations"] = False
        cls._default_config_sources = deepcopy(config.sources)
        return deepcopy(cls._default_config_sources)

    @cached_property
    def config(self) -> beets.IncludeLazyConfig:
        """Reset the shared config to a fresh test baseline."""
        config = beets.config
        config.clear()
        config._materialized = True
        config.sources.extend(self.default_config_sources())
        return config


class RunMixin:
    lib: Library

    def run_command(self, *args: str, lib: Library | None = None) -> None:
        """Run a beets command with an arbitrary amount of arguments. The
        Library` defaults to `self.lib`, but can be overridden with
        the keyword argument `lib`.
        """
        sys.argv = ["beet", *args]  # avoid leakage from test suite args
        lib = lib or self.lib

        with (
            patch.object(lib, "_close", Mock()),
            patch("beets.ui._open_library", return_value=lib),
        ):
            beets.ui._raw_main(list(args))


@pytest.mark.usefixtures("io")
class IOMixin(RunMixin):
    io: _common.DummyIO

    def run_with_output(self, *args: str) -> str:
        self.io.getoutput()
        self.run_command(*args)
        return self.io.getoutput()


class PathsMixin:
    resource_path = _common.RSRC / "full.mp3"

    @cached_property
    def temp_path(self) -> Path:
        return Path(self.create_temp_dir())

    def create_temp_dir(self, **kwargs: Any) -> str:
        return mkdtemp(**kwargs)

    def remove_temp_dir(self) -> None:
        """Delete the temporary directory created by `create_temp_dir`."""
        shutil.rmtree(self.temp_path)


class TestHelper(RunMixin, PathsMixin, ConfigMixin):
    """Helper mixin for high-level cli and plugin tests.

    This mixin provides methods to isolate beets' global state.

    You may use it as a context manager in pytest fixtures in order to setup
    tests at a class or module level. See ``module_helper`` and ``class_helper``
    fixtures, for example.
    """

    request: pytest.FixtureRequest

    def __enter__(self) -> Self:
        self.setup_beets()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> Literal[False]:
        self.teardown_beets()
        # return False/None to propagate exceptions
        return False

    @pytest.fixture(autouse=True)
    def setup(self, request: pytest.FixtureRequest) -> Iterator[None]:
        self.request = request
        self.setup_beets()
        try:
            yield
        finally:
            self.teardown_beets()

    lib: Library

    db_on_disk: ClassVar[bool] = False

    @cached_property
    def lib_path(self) -> Path:
        lib_path = self.temp_path / "libdir"
        lib_path.mkdir(exist_ok=True)
        return lib_path

    # TODO automate teardown through hook registration

    def setup_beets(self) -> None:
        """Setup pristine global configuration and library for testing.

        Sets ``beets.config`` so we can safely use any functionality
        that uses the global configuration.  All paths used are
        contained in a temporary directory

        Sets the following properties on itself.

        - ``temp_path`` Path to a temporary directory containing all
          files specific to beets

        - ``lib_path`` Path to a subfolder of ``temp_path``, containing the
          library's media files. Same as ``config['directory']``.

        - ``lib`` Library instance created with the settings from
          ``config``.

        Make sure you call ``teardown_beets()`` afterwards.
        """
        temp_dir_str = str(self.temp_path)
        self.env_patcher = patch.dict(
            "os.environ",
            {
                "BEETSDIR": temp_dir_str,
                "HOME": temp_dir_str,  # used by Confuse to create directories.
            },
        )
        self.env_patcher.start()

        self.config["directory"] = str(self.lib_path)

        dbpath = (
            self.config["library"].as_path()
            if self.db_on_disk
            else Path(":memory:")
        )
        self.lib = Library(dbpath, str(self.lib_path))

    def teardown_beets(self) -> None:
        self.env_patcher.stop()
        self.lib._close()
        self.remove_temp_dir()

    # Library fixtures methods

    def create_item(self, **values: Any) -> Item:
        """Return an `Item` instance with sensible default values.

        The item receives its attributes from `**values` paratmeter. The
        `title`, `artist`, `album`, `track`, `format` and `path`
        attributes have defaults if they are not given as parameters.
        The `title` attribute is formatted with a running item count to
        prevent duplicates. The default for the `path` attribute
        respects the `format` value.

        The item is attached to the database from `self.lib`.
        """
        values_: dict[str, Any] = {
            "title": "t\u00eftle 1",
            "artist": "the \u00e4rtist",
            "album": "the \u00e4lbum",
            "track": 1,
            "format": "MP3",
        }
        values_.update(values)
        values_["db"] = self.lib
        item = Item(**values_)
        if "path" not in values:
            item["path"] = f"audio.{item['format'].lower()}"
        # mtime needs to be set last since other assignments reset it.
        item.mtime = 12345
        return item

    def add_item(self, **values: Any) -> Item:
        """Add an item to the library and return it.

        Creates the item by passing the parameters to `create_item()`.

        If `path` is not set in `values` it is set to `item.destination()`.
        """
        # When specifying a path, store it normalized (as beets does
        # ordinarily).
        if "path" in values:
            values["path"] = util.normpath(values["path"])

        item = self.create_item(**values)
        item.add(self.lib)

        # Ensure every item has a path.
        if "path" not in values:
            item["path"] = item.destination()
            item.store()

        return item

    def add_item_fixture(self, **values: Any) -> Item:
        """Add an item with an actual audio file to the library."""
        item = self.create_item(**values)
        extension = item["format"].lower()
        item["path"] = _common.RSRC / f"min.{extension}"
        item.add(self.lib)
        item.move(operation=MoveOperation.COPY)
        item.store()
        return item

    def add_album(self, **values: Any) -> Album:
        item = self.add_item(**values)
        return self.lib.add_album([item])

    def add_item_fixtures(self, ext: str = "mp3", count: int = 1) -> list[Item]:
        """Add a number of items with files to the database."""
        # TODO base this on `add_item()`
        items = []
        path = _common.RSRC / f"full.{ext}"
        for i in range(count):
            item = Item.from_path(path)
            item.album = f"\u00e4lbum {i}"  # Check unicode paths
            item.title = f"t\u00eftle {i}"
            # mtime needs to be set last since other assignments reset it.
            item.mtime = 12345
            item.add(self.lib)
            item.move(operation=MoveOperation.COPY)
            item.store()
            items.append(item)
        return items

    def add_album_fixture(
        self,
        track_count: int = 1,
        fname: str = "full",
        ext: str = "mp3",
        disc_count: int = 1,
    ) -> Album:
        """Add an album with files to the database."""
        items = []
        path = _common.RSRC / f"{fname}.{ext}"
        for discnumber in range(1, disc_count + 1):
            for i in range(track_count):
                item = Item.from_path(path)
                item.album = "\u00e4lbum"  # Check unicode paths
                item.title = f"t\u00eftle {i}"
                item.disc = discnumber
                # mtime needs to be set last since other assignments reset it.
                item.mtime = 12345
                item.add(self.lib)
                item.move(operation=MoveOperation.COPY)
                item.store()
                items.append(item)
        return self.lib.add_album(items)

    def create_mediafile_fixture(
        self,
        ext: str = "mp3",
        images: list[str] | None = None,
        target_dir: util.PathLike | None = None,
    ) -> bytes:
        """Copy a fixture mediafile with the extension to `temp_path`.

        `images` is a subset of 'png', 'jpg', and 'tiff'. For each
        specified extension a cover art image is added to the media
        file.
        """
        if not target_dir:
            target_dir = self.temp_path
        src = _common.RSRC / f"full.{ext}"
        handle, path = mkstemp(dir=target_dir)
        path = bytestring_path(path)
        os.close(handle)
        shutil.copyfile(syspath(src), syspath(path))

        if images:
            mediafile = MediaFile(path)
            imgs = []
            for img_ext in images:
                img_path = _common.RSRC / f"image-2x3.{img_ext}"
                with img_path.open("rb") as f:
                    imgs.append(Image(f.read()))
            mediafile.images = imgs
            mediafile.save()

        return path

    # Safe file operations

    def touch(
        self,
        path: util.PathLike,
        dir_: util.PathLike | None = None,
        content: str = "",
    ) -> bytes:
        """Create a file at `path` with given content.

        If `dir_` is given, it is prepended to `path`. After that, if the
        path is relative, it is resolved with respect to
        `self.temp_path`.
        """
        bytes_path = os.fsencode(path)
        if dir_:
            bytes_path = os.path.join(os.fsencode(dir_), bytes_path)

        if not os.path.isabs(bytes_path):
            bytes_path = os.path.join(os.fsencode(self.temp_path), bytes_path)

        parent = os.path.dirname(bytes_path)
        if not os.path.isdir(syspath(parent)):
            os.makedirs(syspath(parent))

        with open(syspath(bytes_path), "a+") as f:
            f.write(content)
        return bytes_path


# A test harness for all beets tests.
# Provides temporary, isolated configuration.
class BeetsTestCase(unittest.TestCase, TestHelper):
    """A unittest.TestCase subclass that saves and restores beets'
    global configuration. This allows tests to make temporary
    modifications that will then be automatically removed when the test
    completes. Also provides some additional assertion methods, a
    temporary directory, and a DummyIO.

    DEPRECATED: Use TestHelper instead.
    """


class ItemInDBTestCase(BeetsTestCase):
    """A test case that includes an in-memory library object (`lib`) and
    an item added to the library (`i`).
    """

    def setUp(self) -> None:
        super().setUp()
        self.i = _common.item(self.lib)


class PytestTestHelper(TestHelper):
    """Same as the BeetsTestCase unittest setup but for pytest."""

    @pytest.fixture(autouse=True)
    def setup(self) -> Iterator[None]:
        self.setup_beets()
        try:
            yield
        finally:
            self.teardown_beets()


class PluginMixin(ConfigMixin):
    plugin: ClassVar[str]
    preload_plugin: ClassVar[bool] = True

    def setup_beets(self) -> None:
        super().setup_beets()  # type: ignore[misc]
        if self.preload_plugin:
            self.load_plugins()

    def teardown_beets(self) -> None:
        super().teardown_beets()  # type: ignore[misc]
        self.unload_plugins()

    def register_plugin(
        self, plugin_class: type[beets.plugins.BeetsPlugin]
    ) -> None:
        beets.plugins._instances.append(plugin_class())

    def load_plugins(self, *plugins: str) -> None:
        """Load and initialize plugins by names.

        Similar setting a list of plugins in the configuration. Make
        sure you call ``unload_plugins()`` afterwards.
        """
        # FIXME this should eventually be handled by a plugin manager
        plugins = (self.plugin,) if hasattr(self, "plugin") else plugins
        self.config["plugins"] = plugins
        beets.plugins.load_plugins()

    def unload_plugins(self) -> None:
        """Unload all plugins and remove them from the configuration."""
        # FIXME this should eventually be handled by a plugin manager
        beets.plugins.BeetsPlugin.listeners.clear()
        beets.plugins.BeetsPlugin._raw_listeners.clear()
        self.config["plugins"] = []
        beets.plugins._instances.clear()

    @contextmanager
    def configure_plugin(self, config: Any) -> Iterator[None]:
        self.config[self.plugin].set(config)
        self.load_plugins(self.plugin)

        yield

        self.unload_plugins()


class PluginTestCase(PluginMixin, BeetsTestCase):
    """
    DEPRECATED: Use PluginTestHelper instead.
    """


class PluginTestHelper(PluginMixin, TestHelper):
    """Helper mixin for pytest-based plugin tests.

    This mixin provides the standard beets test setup and automatically
    initializes and tears down plugin state for each test.

    .. code-block:: python

        class TestMyPlugin(PluginTestHelper):
            plugin: ClassVar[str] = "myplugin"
    """


class ImporterMixin(PathsMixin, ConfigMixin):
    """Provides tools to setup a library, a directory containing files that are
    to be imported and an import session. The class also provides stubs for the
    autotagging library and several assertions for the library.
    """

    default_import_config: ClassVar[dict[str, Any]] = {
        "autotag": True,
        "copy": True,
        "hardlink": False,
        "link": False,
        "move": False,
        "resume": False,
        "singletons": False,
        "timid": True,
    }

    lib: Library
    importer: ImportSession
    import_media: list[MediaFile]

    @cached_property
    def import_path(self) -> Path:
        import_path = self.temp_path / "import"
        import_path.mkdir(exist_ok=True)
        return import_path

    def prepare_track_for_import(
        self, track_id: int, album_path: Path, album_id: int | None = None
    ) -> Path:
        track_path = album_path / f"track_{track_id}.mp3"
        shutil.copy(self.resource_path, track_path)
        medium = MediaFile(track_path)
        medium.update(
            {
                "album": f"Tag Album{f' {album_id}' if album_id else ''}",
                "albumartist": None,
                "mb_albumid": None,
                "comp": None,
                "artist": "Tag Artist",
                "title": f"Tag Track {track_id}",
                "track": track_id,
                "mb_trackid": None,
            }
        )
        medium.save()
        self.import_media.append(medium)
        return track_path

    def prepare_album_for_import(
        self,
        item_count: int,
        album_id: int | None = None,
        album_path: Path | None = None,
    ) -> list[Path]:
        """Create an album directory with media files to import.

        The directory has following layout
          album/
            track_1.mp3
            track_2.mp3
            track_3.mp3
        """
        if not album_path:
            album_dir = f"album_{album_id}" if album_id else "album"
            album_path = self.import_path / album_dir

        album_path.mkdir(exist_ok=True)

        return [
            self.prepare_track_for_import(tid, album_path, album_id=album_id)
            for tid in range(1, item_count + 1)
        ]

    def prepare_albums_for_import(self, count: int = 1) -> None:
        album_dirs = self.import_path.glob("album_*")
        base_idx = int(str(max(album_dirs, default="0")).split("_")[-1]) + 1

        for album_id in range(base_idx, count + base_idx):
            self.prepare_album_for_import(1, album_id=album_id)

    def _get_import_session(self, import_dir: Path) -> ImportSession:
        return ImportSessionFixture(
            self.lib,
            loghandler=None,
            query=None,
            paths=[os.fsencode(import_dir)],
        )

    def setup_importer(
        self, import_dir: Path | None = None, **kwargs: Any
    ) -> ImportSession:
        self.config["import"].set_args({**self.default_import_config, **kwargs})
        self.importer = self._get_import_session(import_dir or self.import_path)
        return self.importer

    def setup_singleton_importer(self, **kwargs: Any) -> ImportSession:
        return self.setup_importer(singletons=True, **kwargs)


class ImportHelper(TestHelper, ImporterMixin):
    def setup_beets(self) -> None:
        super().setup_beets()
        self.import_media = []
        self.lib.path_formats = [
            ("default", template(os.path.join("$artist", "$album", "$title"))),
            ("singleton:true", template(os.path.join("singletons", "$title"))),
            (
                "comp:true",
                template(os.path.join("compilations", "$album", "$title")),
            ),
        ]


class AsIsImporterMixin(ImporterMixin):
    def setup_beets(self) -> None:
        super().setup_beets()  # type: ignore[misc]
        self.prepare_album_for_import(1)

    def run_asis_importer(self, **kwargs: Any) -> ImportSession:
        importer = self.setup_importer(autotag=False, **kwargs)
        importer.run()
        return importer


class ImportSessionFixture(ImportSession):
    """ImportSession that can be controlled programaticaly.

    >>> lib = Library(':memory:')
    >>> importer = ImportSessionFixture(lib, paths=['/path/to/import'])
    >>> importer.add_choice(importer.Action.SKIP)
    >>> importer.add_choice(importer.Action.ASIS)
    >>> importer.default_choice = importer.Action.APPLY
    >>> importer.run()

    This imports ``/path/to/import`` into `lib`. It skips the first
    album and imports the second one with metadata from the tags. For the
    remaining albums, the metadata from the autotagger will be applied.
    """

    _choices: list[importer.Action | int]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._choices = []

    default_choice = importer.Action.APPLY

    def add_choice(self, choice: importer.Action | int) -> None:
        self._choices.append(choice)

    def clear_choices(self) -> None:
        self._choices = []

    def choose_match(
        self, task: importer.ImportTask
    ) -> AlbumMatch | importer.Action:
        try:
            choice = self._choices.pop(0)
        except IndexError:
            choice = self.default_choice

        if task.candidates:
            if choice == importer.Action.APPLY:
                return task.candidates[0]  # type: ignore[return-value]
            if isinstance(choice, int):
                return task.candidates[choice - 1]  # type: ignore[return-value]

        assert not isinstance(choice, int), f"Invalid choice: {choice}"
        return choice

    choose_item = choose_match  # type: ignore[assignment]


class TerminalImportSessionFixture(TerminalImportSession):
    def __init__(self, *args, **kwargs):
        self.io = kwargs.pop("io")
        super().__init__(*args, **kwargs)
        self._choices = []

    default_choice = importer.Action.APPLY

    def add_choice(self, choice: importer.Action | int) -> None:
        self._choices.append(choice)

    def clear_choices(self) -> None:
        self._choices = []

    def choose_match(
        self, task: importer.ImportTask
    ) -> AlbumMatch | importer.Action:
        self._add_choice_input()
        return super().choose_match(task)

    def choose_item(
        self, task: importer.ImportTask
    ) -> TrackMatch | importer.Action:
        self._add_choice_input()
        return super().choose_item(task)

    def _add_choice_input(self) -> None:
        try:
            choice = self._choices.pop(0)
        except IndexError:
            choice = self.default_choice

        if choice == importer.Action.APPLY:
            self.io.addinput("A")
        elif choice == importer.Action.ASIS:
            self.io.addinput("U")
        elif choice == importer.Action.ALBUMS:
            self.io.addinput("G")
        elif choice == importer.Action.TRACKS:
            self.io.addinput("T")
        elif choice == importer.Action.SKIP:
            self.io.addinput("S")
        else:
            self.io.addinput("M")
            self.io.addinput(str(choice))
            self._add_choice_input()


class TerminalImportMixin(IOMixin, ImportHelper):
    """Provides_a terminal importer for the import session."""

    def _get_import_session(self, import_dir: Path) -> importer.ImportSession:
        return TerminalImportSessionFixture(
            self.lib,
            loghandler=None,
            query=None,
            io=self.request.getfixturevalue("io"),
            paths=[os.fsencode(import_dir)],
        )


@dataclass
class AutotagStub:
    """Stub out MusicBrainz album and track matcher and control what the
    autotagger returns.
    """

    NONE = "NONE"
    IDENT = "IDENT"
    GOOD = "GOOD"
    BAD = "BAD"
    MISSING = "MISSING"
    matching: str

    length = 2

    def install(self) -> Self:
        self.patchers: list[_patch[Any]] = [
            patch("beets.metadata_plugins.album_for_id", lambda *_: None),
            patch("beets.metadata_plugins.track_for_id", lambda *_: None),
            patch("beets.metadata_plugins.candidates", self.candidates),
            patch(
                "beets.metadata_plugins.item_candidates", self.item_candidates
            ),
        ]
        for p in self.patchers:
            p.start()

        return self

    def restore(self) -> None:
        for p in self.patchers:
            p.stop()

    def candidates(
        self, items: Sequence[Item], artist: str, album: str, _: bool
    ) -> Iterable[AlbumInfo]:
        if self.matching == self.IDENT:
            yield self._make_album_match(artist, album, len(items))

        elif self.matching == self.GOOD:
            for i in range(self.length):
                yield self._make_album_match(artist, album, len(items), i)

        elif self.matching == self.BAD:
            for i in range(self.length):
                yield self._make_album_match(artist, album, len(items), i + 1)

        elif self.matching == self.MISSING:
            yield self._make_album_match(artist, album, len(items), missing=1)

    def item_candidates(
        self, item: Item, artist: str, title: str
    ) -> Iterable[TrackInfo]:
        yield TrackInfo(
            title=title.replace("Tag", "Applied"),
            track_id="trackid",
            artist=artist.replace("Tag", "Applied"),
            artist_id="artistid",
            length=1,
            index=0,
        )

    def _make_track_match(
        self, artist: str, album: str, number: int
    ) -> TrackInfo:
        return TrackInfo(
            title=f"Applied Track {number}",
            track_id=f"match {number}",
            artist=artist,
            length=1,
            index=0,
        )

    def _make_album_match(
        self,
        artist: str,
        album: str,
        tracks: int,
        distance: int = 0,
        missing: int = 0,
    ) -> AlbumInfo:
        id_ = f" {'M' * distance}" if distance else ""

        artist = f"{artist.replace('Tag', 'Applied')}{id_}"
        album = f"{album.replace('Tag', 'Applied')}{id_}"

        track_infos = []
        for i in range(tracks - missing):
            track_infos.append(self._make_track_match(artist, album, i + 1))

        return AlbumInfo(
            artist=artist,
            album=album,
            tracks=track_infos,
            va=False,
            album_id=f"albumid{id_}",
            artist_id=f"artistid{id_}",
            albumtype="soundtrack",
            data_source="match_source",
            bandcamp_album_id="bc_url",
        )


class AutotagImportHelper(ImportHelper):
    matching = AutotagStub.IDENT

    def setup_beets(self) -> None:
        super().setup_beets()
        self.matcher = AutotagStub(self.matching).install()

    def teardown_beets(self) -> None:
        self.matcher.restore()
        super().teardown_beets()


class AutotagImportTestCase(AutotagImportHelper, BeetsTestCase):
    """DEPRECATED: Use AutotagImportHelper instead."""


@dataclass(slots=True)
class ImageRequestMocker:
    mocker: Mocker

    # Image types and their file headers
    IMAGE_HEADERS: ClassVar[dict[str, bytes]] = {
        "image/jpeg": b"\xff\xd8\xff\x00\x00\x00JFIF",
        "image/png": b"\211PNG\r\n\032\n",
        "image/gif": b"GIF89a",
        # dummy type that is definitely not a valid image content type
        "image/watercolour": b"watercolour",
        "text/html": (
            b"<!DOCTYPE html>\n<html>\n<head>\n</head>\n"
            b"<body>\n</body>\n</html>"
        ),
    }

    def get(
        self,
        url: str,
        *,
        content_type: str = "image/jpeg",
        file_type: str | None = None,
        content: str | bytes | None = None,
    ) -> None:
        actual_file_type = file_type or content_type

        if content is None:
            try:
                content = self.IMAGE_HEADERS[actual_file_type].ljust(
                    32, b"\x00"
                )
            except KeyError as exc:
                # If we can't return a file that looks like real file of the requested
                # type, better fail the test than returning something else, which might
                # violate assumption made when writing a test.
                raise AssertionError(
                    f"Mocking {actual_file_type!r} responses not supported"
                ) from exc

        if isinstance(content, str):
            content = content.encode()

        self.mocker.get(
            url, headers={"Content-Type": content_type}, content=content
        )


class FetchImageHelper:
    """Pytest mixin providing image response mocking utilities."""

    @pytest.fixture
    def image_request_mock(self, requests_mock: Mocker) -> ImageRequestMocker:
        return ImageRequestMocker(requests_mock)


class CleanupModulesMixin:
    modules: ClassVar[tuple[str, ...]]

    @classmethod
    def tearDownClass(cls) -> None:
        """Remove files created by the plugin."""
        for module in cls.modules:
            clean_module_tempdir(module)



"""Tests for the 'ftintitle' plugin."""

from __future__ import annotations

from typing import cast

import pytest

from beets import plugins
from beets.autotag import AlbumInfo, TrackInfo
from beets.importer import ImportSession, SingletonImportTask
from beets.library.models import Album
from beets.test.helper import PluginTestHelper
from beetsplug import ftintitle


class TestFtInTitlePluginFunctional(PluginTestHelper):
    plugin = "ftintitle"
    preload_plugin = False

    @pytest.mark.parametrize(
        "cfg, cmd_args, given, expected",
        [
            pytest.param(
                {},
                ("ftintitle",),
                ("Alice", "Song 1", "Alice"),
                ("Alice", "Song 1"),
                id="no-featured-artist",
            ),
            pytest.param(
                {"format": "feat {0}"},
                ("ftintitle",),
                ("Alice ft. Bob", "Song 1", None),
                ("Alice", "Song 1 feat Bob"),
                id="no-albumartist-custom-format",
            ),
            pytest.param(
                {},
                ("ftintitle",),
                ("Alice", "Song 1", None),
                ("Alice", "Song 1"),
                id="no-albumartist-no-feature",
            ),
            pytest.param(
                {"format": "featuring {0}"},
                ("ftintitle",),
                ("Alice ft Bob", "Song 1", "George"),
                ("Alice", "Song 1 featuring Bob"),
                id="guest-artist-custom-format",
            ),
            pytest.param(
                {},
                ("ftintitle",),
                ("Alice", "Song 1", "George"),
                ("Alice", "Song 1"),
                id="guest-artist-no-feature",
            ),
            # ---- drop (-d) variants ----
            pytest.param(
                {},
                ("ftintitle", "-d"),
                ("Alice ft Bob", "Song 1", "Alice"),
                ("Alice", "Song 1"),
                id="drop-self-ft",
            ),
            pytest.param(
                {},
                ("ftintitle", "-d"),
                ("Alice", "Song 1", "Alice"),
                ("Alice", "Song 1"),
                id="drop-self-no-ft",
            ),
            pytest.param(
                {},
                ("ftintitle", "-d"),
                ("Alice ft Bob", "Song 1", "George"),
                ("Alice", "Song 1"),
                id="drop-guest-ft",
            ),
            pytest.param(
                {},
                ("ftintitle", "-d"),
                ("Alice", "Song 1", "George"),
                ("Alice", "Song 1"),
                id="drop-guest-no-ft",
            ),
            # ---- custom format variants ----
            pytest.param(
                {"format": "feat. {}"},
                ("ftintitle",),
                ("Alice ft Bob", "Song 1", "Alice"),
                ("Alice", "Song 1 feat. Bob"),
                id="custom-format-feat-dot",
            ),
            pytest.param(
                {"format": "featuring {}"},
                ("ftintitle",),
                ("Alice feat. Bob", "Song 1", "Alice"),
                ("Alice", "Song 1 featuring Bob"),
                id="custom-format-featuring",
            ),
            pytest.param(
                {"format": "with {}"},
                ("ftintitle",),
                ("Alice feat Bob", "Song 1", "Alice"),
                ("Alice", "Song 1 with Bob"),
                id="custom-format-with",
            ),
            pytest.param(
                {"format": "ft. {}"},
                ("ftintitle",),
                ("Alice (ft. Bob)", "Song 1", "Alice"),
                ("Alice", "Song 1 ft. Bob"),
                id="featured-artist-in-parentheses",
            ),
            # ---- keep_in_artist variants ----
            pytest.param(
                {"format": "feat. {}", "keep_in_artist": True},
                ("ftintitle",),
                ("Alice ft Bob", "Song 1", "Alice"),
                ("Alice ft Bob", "Song 1 feat. Bob"),
                id="keep-in-artist-add-to-title",
            ),
            pytest.param(
                {"format": "feat. {}", "keep_in_artist": True},
                ("ftintitle", "-d"),
                ("Alice ft Bob", "Song 1", "Alice"),
                ("Alice ft Bob", "Song 1"),
                id="keep-in-artist-drop-from-title",
            ),
            # ---- custom_words variants ----
            pytest.param(
                {"format": "featuring {}", "custom_words": ["med"]},
                ("ftintitle",),
                ("Alice med Bob", "Song 1", "Alice"),
                ("Alice", "Song 1 featuring Bob"),
                id="custom-feat-words",
            ),
            pytest.param(
                {
                    "format": "featuring {}",
                    "keep_in_artist": True,
                    "custom_words": ["med"],
                },
                ("ftintitle",),
                ("Alice med Bob", "Song 1", "Alice"),
                ("Alice med Bob", "Song 1 featuring Bob"),
                id="custom-feat-words-keep-in-artists",
            ),
            pytest.param(
                {
                    "format": "featuring {}",
                    "keep_in_artist": True,
                    "custom_words": ["med"],
                },
                ("ftintitle", "-d"),
                ("Alice med Bob", "Song 1", "Alice"),
                ("Alice med Bob", "Song 1"),
                id="custom-feat-words-keep-in-artists-drop-from-title",
            ),
            # ---- preserve_album_artist variants ----
            pytest.param(
                {"format": "feat. {}", "preserve_album_artist": True},
                ("ftintitle",),
                ("Alice feat. Bob", "Song 1", "Alice"),
                ("Alice", "Song 1 feat. Bob"),
                id="skip-if-artist-and-album-artists-is-the-same-different-match",
            ),
            pytest.param(
                {"format": "feat. {}", "preserve_album_artist": False},
                ("ftintitle",),
                ("Alice feat. Bob", "Song 1", "Alice"),
                ("Alice", "Song 1 feat. Bob"),
                id="skip-if-artist-and-album-artists-is-the-same-different-match-b",
            ),
            pytest.param(
                {"format": "feat. {}", "preserve_album_artist": True},
                ("ftintitle",),
                ("Alice feat. Bob", "Song 1", "Alice feat. Bob"),
                ("Alice feat. Bob", "Song 1"),
                id="skip-if-artist-and-album-artists-is-the-same-matching-match",
            ),
            pytest.param(
                {"format": "feat. {}", "preserve_album_artist": False},
                ("ftintitle",),
                ("Alice feat. Bob", "Song 1", "Alice feat. Bob"),
                ("Alice", "Song 1 feat. Bob"),
                id="skip-if-artist-and-album-artists-is-the-same-matching-match-b",
            ),
            # ---- titles with brackets/parentheses ----
            pytest.param(
                {"format": "ft. {}", "bracket_keywords": ["mix"]},
                ("ftintitle",),
                ("Alice ft. Bob", "Song 1 (Club Mix)", "Alice"),
                ("Alice", "Song 1 ft. Bob (Club Mix)"),
                id="ft-inserted-before-matching-bracket-keyword",
            ),
            pytest.param(
                {"format": "ft. {}", "bracket_keywords": ["nomatch"]},
                ("ftintitle",),
                ("Alice ft. Bob", "Song 1 (Club Remix)", "Alice"),
                ("Alice", "Song 1 (Club Remix) ft. Bob"),
                id="ft-inserted-at-end-no-bracket-keyword-match",
            ),
        ],
    )
    def test_ftintitle_functional(
        self,
        cfg: dict[str, str | bool | list[str]],
        cmd_args: tuple[str, ...],
        given: tuple[str, str, str | None],
        expected: tuple[str, str],
    ) -> None:
        config = {
            "drop": False,
            "auto": True,
            "keep_in_artist": False,
            "custom_words": [],
            **cfg,
        }

        artist, title, albumartist = given
        item = self.add_item(
            path="/",
            artist=artist,
            artist_sort=artist,
            title=title,
            albumartist=albumartist,
        )

        with self.configure_plugin(config):
            self.run_command(*cmd_args)

        item.load()

        expected_artist, expected_title = expected
        assert item["artist"] == expected_artist
        assert item["title"] == expected_title

    def test_trackinfo_received_rewrites_artist_when_artist_credit_enabled(
        self,
    ) -> None:
        self.config["artist_credit"] = True
        info = TrackInfo(
            artist="Alice feat. Bob",
            artist_credit="Alice",
            artist_sort="Alice feat. Bob",
            artists=["Alice", "Bob"],
            artists_credit=["Alice"],
            title="Song",
        )

        with self.configure_plugin({"auto": True}):
            plugins.send("trackinfo_received", info=info)

        assert info.artist == "Alice"
        assert info.artist_credit == "Alice"
        assert info.artist_sort == "Alice"
        assert info.artists == ["Alice", "Bob"]
        assert info.artists_credit == ["Alice"]
        assert info.title == "Song feat. Bob"
        assert info.item_data["title"] == "Song feat. Bob"

    @pytest.mark.parametrize(
        "config, artist, title, expected_artist, expected_title",
        [
            pytest.param(
                {"auto": False},
                "Alice feat. Bob",
                "Song",
                "Alice feat. Bob",
                "Song",
                id="auto-disabled",
            ),
            pytest.param(
                {"auto": True},
                "Alice",
                "Song",
                "Alice",
                "Song",
                id="no-feature",
            ),
            pytest.param(
                {"auto": True},
                "Alice feat. Bob",
                "Song feat. Bob",
                "Alice",
                "Song feat. Bob",
                id="existing-title-feature",
            ),
        ],
    )
    def test_trackinfo_received_rewrites_or_skips_simple_cases(
        self,
        config: dict[str, bool],
        artist: str,
        title: str,
        expected_artist: str,
        expected_title: str,
    ) -> None:
        info = TrackInfo(artist=artist, title=title)

        with self.configure_plugin(config):
            plugins.send("trackinfo_received", info=info)

        assert info.artist == expected_artist
        assert info.title == expected_title

    def test_ft_in_title_reports_no_change_for_info_drop_keep_noop(
        self,
    ) -> None:
        info = TrackInfo(artist="Alice feat. Bob", title="Song")

        with self.configure_plugin({"drop": True, "keep_in_artist": True}):
            plugin = next(iter(plugins.find_plugins()))
            assert isinstance(plugin, ftintitle.FtInTitlePlugin)
            assert plugin.ft_in_title(info, "") is False

        assert info.artist == "Alice feat. Bob"
        assert info.title == "Song"

    def test_trackinfo_received_preserves_artist_credit_when_disabled(
        self,
    ) -> None:
        self.config["artist_credit"] = False
        info = TrackInfo(
            artist="Alice feat. Bob",
            artist_credit="Alice feat. Bobby",
            title="Song",
        )

        with self.configure_plugin({"auto": True}):
            plugins.send("trackinfo_received", info=info)

        assert info.artist == "Alice"
        assert info.artist_credit == "Alice feat. Bobby"
        assert info.title == "Song feat. Bob"
        assert info.item_data["artist_credit"] == "Alice feat. Bobby"

    def test_ft_in_title_reports_no_change_for_item_drop_keep_noop(
        self,
    ) -> None:
        item = self.add_item(
            path="/",
            artist="Alice feat. Bob",
            artist_sort="",
            title="Song",
            albumartist="Alice",
        )

        with self.configure_plugin({"drop": True, "keep_in_artist": True}):
            plugin = next(iter(plugins.find_plugins()))
            assert isinstance(plugin, ftintitle.FtInTitlePlugin)
            assert (
                plugin.ft_in_title(item, item.get("albumartist") or "") is False
            )

        assert item.artist == "Alice feat. Bob"
        assert item.title == "Song"

    def test_imported_stage_rewrites_imported_items(self) -> None:
        item = self.add_item(
            path="/",
            artist="Alice feat. Bob",
            title="Song",
            albumartist="Alice",
        )

        with self.configure_plugin({"auto": True}):
            plugin = next(iter(plugins.find_plugins()))
            assert isinstance(plugin, ftintitle.FtInTitlePlugin)
            plugin.imported(
                cast(ImportSession, None), SingletonImportTask(None, item)
            )

        item.load()
        assert item.artist == "Alice"
        assert item.title == "Song feat. Bob"

    def test_albuminfo_received_rewrites_tracks_with_album_artist(self) -> None:
        track_info = TrackInfo(
            artist="Alice & Bob", artist_sort="Alice & Bob", title="Song"
        )
        info = AlbumInfo(artist="Alice", album="Album", tracks=[track_info])

        with self.configure_plugin({"auto": True}):
            plugins.send("albuminfo_received", info=info)

        assert track_info.artist == "Alice"
        assert track_info.artist_sort == "Alice"
        assert track_info.title == "Song feat. Bob"
        assert track_info.item_data["artist"] == "Alice"
        assert track_info.item_data["title"] == "Song feat. Bob"


@pytest.mark.parametrize(
    "artist,albumartist,expected",
    [
        ("Alice ft. Bob", "Alice", "Bob"),
        ("Alice feat Bob", "Alice", "Bob"),
        ("Alice featuring Bob", "Alice", "Bob"),
        ("Alice & Bob", "Alice", "Bob"),
        ("Alice and Bob", "Alice", "Bob"),
        ("Alice With Bob", "Alice", "Bob"),
        ("Alice (ft. Bob)", "Alice", "Bob"),
        ("Alice [ft. Bob]", "Bob", "Alice"),
        ("Alice defeat Bob", "Alice", None),
        ("Alice & Bob", "Bob", "Alice"),
        ("Alice ft. Bob", "Bob", "Alice"),
        ("Alice ft. Carol", "Bob", "Carol"),
    ],
)
def test_find_feat_part(
    artist: str, albumartist: str, expected: str | None
) -> None:
    assert ftintitle.find_feat_part(artist, albumartist) == expected


@pytest.mark.parametrize(
    "given,expected",
    [
        ("Alice ft. Bob", ("Alice", "Bob")),
        ("Alice feat Bob", ("Alice", "Bob")),
        ("Alice feat. Bob", ("Alice", "Bob")),
        ("Alice featuring Bob", ("Alice", "Bob")),
        ("Alice & Bob", ("Alice", "Bob")),
        ("Alice, Bob & Charlie", ("Alice", "Bob & Charlie")),
        (
            "Alice, Bob & Charlie feat. Xavier",
            ("Alice, Bob & Charlie", "Xavier"),
        ),
        ("Alice and Bob", ("Alice", "Bob")),
        ("Alice With Bob", ("Alice", "Bob")),
        ("Alice defeat Bob", ("Alice defeat Bob", None)),
        ("Alice & Bob feat Charlie", ("Alice & Bob", "Charlie")),
        ("Alice & Bob ft. Charlie", ("Alice & Bob", "Charlie")),
        ("Alice & Bob featuring Charlie", ("Alice & Bob", "Charlie")),
        ("Alice and Bob feat Charlie", ("Alice and Bob", "Charlie")),
        ("Alice (ft. Bob)", ("Alice", "Bob")),
        ("Alice [& Bob]", ("Alice", "Bob")),
    ],
)
def test_split_on_feat(given: str, expected: tuple[str, str | None]) -> None:
    assert ftintitle.split_on_feat(given) == expected


@pytest.mark.parametrize(
    "given,keywords,expected",
    [
        ## default keywords
        # different braces and keywords
        ("Song (Remix)", None, "Song ft. Bob (Remix)"),
        ("Song [Version]", None, "Song ft. Bob [Version]"),
        ("Song {Extended Mix}", None, "Song ft. Bob {Extended Mix}"),
        ("Song <Instrumental>", None, "Song ft. Bob <Instrumental>"),
        # two keyword clauses
        ("Song (Remix) (Live)", None, "Song ft. Bob (Remix) (Live)"),
        # brace insensitivity
        ("Song (Live) [Remix]", None, "Song ft. Bob (Live) [Remix]"),
        ("Song [Edit] (Remastered)", None, "Song ft. Bob [Edit] (Remastered)"),
        # negative cases
        ("Song", None, "Song ft. Bob"),  # no clause
        ("Song (Arbitrary)", None, "Song (Arbitrary) ft. Bob"),  # no keyword
        ("Song (", None, "Song ( ft. Bob"),  # no matching brace or keyword
        ("Song (Live", None, "Song (Live ft. Bob"),  # no matching brace with keyword
        # one keyword clause, one non-keyword clause
        ("Song (Live) (Arbitrary)", None, "Song ft. Bob (Live) (Arbitrary)"),
        ("Song (Arbitrary) (Remix)", None, "Song (Arbitrary) ft. Bob (Remix)"),
        # nested brackets - same type
        ("Song (Remix (Extended))", None, "Song ft. Bob (Remix (Extended))"),
        ("Song [Arbitrary [Description]]", None, "Song [Arbitrary [Description]] ft. Bob"),
        # nested brackets - different types
        ("Song (Remix [Extended])", None, "Song ft. Bob (Remix [Extended])"),
        # nested - returns outer start position despite inner keyword
        ("Song [Arbitrary {Extended}]", None, "Song ft. Bob [Arbitrary {Extended}]"),
        ("Song {Live <Arbitrary>}", None, "Song ft. Bob {Live <Arbitrary>}"),
        ("Song <Remaster (Arbitrary)>", None, "Song ft. Bob <Remaster (Arbitrary)>"),
        ("Song <Extended> [Live]", None, "Song ft. Bob <Extended> [Live]"),
        ("Song (Version) <Live>", None, "Song ft. Bob (Version) <Live>"),
        ("Song (Arbitrary [Description])", None, "Song (Arbitrary [Description]) ft. Bob"),
        ("Song [Description (Arbitrary)]", None, "Song [Description (Arbitrary)] ft. Bob"),
        ## custom keywords
        ("Song (Live)", ["live"], "Song ft. Bob (Live)"),
        ("Song (Concert)", ["concert"], "Song ft. Bob (Concert)"),
        ("Song (Remix)", ["custom"], "Song (Remix) ft. Bob"),
        ("Song (Custom)", ["custom"], "Song ft. Bob (Custom)"),
        ("Song", [], "Song ft. Bob"),
        ("Song (", [], "Song ( ft. Bob"),
        # Multi-word keyword tests
        ("Song (Club Mix)", ["club mix"], "Song ft. Bob (Club Mix)"),  # Positive: matches multi-word
        ("Song (Club Remix)", ["club mix"], "Song (Club Remix) ft. Bob"),  # Negative: no match
    ],
)  # fmt: skip
def test_insert_ft_into_title(
    given: str, keywords: list[str] | None, expected: str
) -> None:
    assert (
        ftintitle.FtInTitlePlugin.insert_ft_into_title(
            given, "ft. Bob", keywords
        )
        == expected
    )


@pytest.mark.parametrize(
    "given,expected",
    [
        ("Alice ft. Bob", True),
        ("Alice feat. Bob", True),
        ("Alice feat Bob", True),
        ("Alice featuring Bob", True),
        ("Alice (ft. Bob)", True),
        ("Alice (feat. Bob)", True),
        ("Alice [ft. Bob]", True),
        ("Alice [feat. Bob]", True),
        ("Alice defeat Bob", False),
        ("Aliceft.Bob", False),
        ("Alice (defeat Bob)", False),
        ("Live and Let Go", False),
        ("Come With Me", False),
    ],
)
def test_contains_feat(given: str, expected: bool) -> None:
    assert ftintitle.contains_feat(given) is expected


@pytest.mark.parametrize(
    "given,custom_words,expected",
    [
        ("Alice ft. Bob", [], True),
        ("Alice feat. Bob", [], True),
        ("Alice feat Bob", [], True),
        ("Alice featuring Bob", [], True),
        ("Alice (ft. Bob)", [], True),
        ("Alice (feat. Bob)", [], True),
        ("Alice [ft. Bob]", [], True),
        ("Alice [feat. Bob]", [], True),
        ("Alice defeat Bob", [], False),
        ("Aliceft.Bob", [], False),
        ("Alice (defeat Bob)", [], False),
        ("Live and Let Go", [], False),
        ("Come With Me", [], False),
        ("Alice x Bob", ["x"], True),
        ("Alice x Bob", ["X"], True),
        ("Alice och Xavier", ["x"], False),
        ("Alice ft. Xavier", ["x"], True),
        ("Alice med Carol", ["med"], True),
        ("Alice med Carol", [], False),
    ],
)
def test_custom_words(
    given: str, custom_words: list[str], expected: bool
) -> None:
    assert ftintitle.contains_feat(given, custom_words) is expected


def test_album_template_value(config):
    config["ftintitle"]["custom_words"] = []

    album = Album()
    album["albumartist"] = "Foo ft. Bar"
    assert ftintitle._album_artist_no_feat(album) == "Foo"

    album["albumartist"] = "Foobar"
    assert ftintitle._album_artist_no_feat(album) == "Foobar"



"""Tests for the 'web' plugin"""

import json
import os.path
import platform
import shutil
from collections import Counter

import pytest

from beets.library import Album, Item
from beets.test import _common
from beets.test.helper import PluginMixin, PytestTestHelper
from beetsplug import web


class WebPluginMixin(PluginMixin):
    """Mixin to configure the web plugin for testing."""

    plugin = "web"

    @pytest.fixture(autouse=True)
    def setup_web_app(self, setup):
        """Configure the web plugin's Flask app for testing.

        This fixture sets up the Flask test client and configures the app
        with the test library. It runs after the base setup fixture from
        PytestTestHelper.
        """
        # Set up the web app config - we modify self.config["web"] for beets config
        # but also need to directly configure the Flask app for tests
        web.app.config["TESTING"] = True
        web.app.config["lib"] = self.lib
        web.app.config["INCLUDE_PATHS"] = False
        web.app.config["READONLY"] = True
        self.client = web.app.test_client()

        # Set platform-specific path prefix
        if platform.system() == "Windows":
            self.path_prefix = "C:"
        else:
            self.path_prefix = ""

        # Add library elements. Note that self.lib.add overrides any "id=<n>"
        # and assigns the next free id number.
        # The following adds will create items #1, #2 and #3
        path1 = (
            self.path_prefix + os.sep + os.path.join(b"path_1").decode("utf-8")
        )
        self.lib.add(
            Item(title="title", path=path1, album_id=2, artist="AAA Singers")
        )
        path2 = (
            self.path_prefix
            + os.sep
            + os.path.join(b"somewhere", b"a").decode("utf-8")
        )
        self.lib.add(
            Item(title="another title", path=path2, artist="AAA Singers")
        )
        path3 = (
            self.path_prefix
            + os.sep
            + os.path.join(b"somewhere", b"abc").decode("utf-8")
        )
        self.lib.add(
            Item(title="and a third", testattr="ABC", path=path3, album_id=2)
        )
        # The following adds will create albums #1 and #2
        self.lib.add(Album(album="album", albumtest="xyz"))
        path4 = (
            self.path_prefix
            + os.sep
            + os.path.join(b"somewhere2", b"art_path_2").decode("utf-8")
        )
        self.lib.add(Album(album="other album", artpath=path4))

        return


class TestWebPlugin(WebPluginMixin, PytestTestHelper):
    """Tests for the web plugin."""

    def test_config_include_paths_true(self):
        web.app.config["INCLUDE_PATHS"] = True
        response = self.client.get("/item/1")
        res_json = json.loads(response.data.decode("utf-8"))
        expected_path = (
            self.path_prefix + os.sep + os.path.join(b"path_1").decode("utf-8")
        )

        assert response.status_code == 200
        assert res_json["path"] == expected_path

        web.app.config["INCLUDE_PATHS"] = False

    def test_config_include_artpaths_true(self):
        web.app.config["INCLUDE_PATHS"] = True
        response = self.client.get("/album/2")
        res_json = json.loads(response.data.decode("utf-8"))
        expected_path = (
            self.path_prefix
            + os.sep
            + os.path.join(b"somewhere2", b"art_path_2").decode("utf-8")
        )

        assert response.status_code == 200
        assert res_json["artpath"] == expected_path

        web.app.config["INCLUDE_PATHS"] = False

    def test_config_include_paths_false(self):
        web.app.config["INCLUDE_PATHS"] = False
        response = self.client.get("/item/1")
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert "path" not in res_json

    def test_config_include_artpaths_false(self):
        web.app.config["INCLUDE_PATHS"] = False
        response = self.client.get("/album/2")
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert "artpath" not in res_json

    def test_get_all_items(self):
        response = self.client.get("/item/")
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert len(res_json["items"]) == 3

    def test_get_unique_item_artist(self):
        response = self.client.get("/item/values/artist")
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert res_json["values"] == ["", "AAA Singers"]

    def test_get_single_item_by_id(self):
        response = self.client.get("/item/1")
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert res_json["id"] == 1
        assert res_json["title"] == "title"

    def test_get_multiple_items_by_id(self):
        response = self.client.get("/item/1,2")
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert len(res_json["items"]) == 2
        response_titles = {item["title"] for item in res_json["items"]}
        assert response_titles == {"title", "another title"}

    def test_get_single_item_not_found(self):
        response = self.client.get("/item/4")
        assert response.status_code == 404

    def test_get_single_item_by_path(self):
        data_path = _common.RSRC / "full.mp3"
        self.lib.add(Item.from_path(data_path))
        response = self.client.get(f"/item/path/{data_path}")
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert res_json["title"] == "full"

    def test_get_single_item_by_path_not_found_if_not_in_library(self):
        data_path = _common.RSRC / "full.mp3"
        # data_path points to a valid file, but we have not added the file
        # to the library.
        response = self.client.get(f"/item/path/{data_path}")

        assert response.status_code == 404

    def test_get_item_empty_query(self):
        response = self.client.get("/item/query/")
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert len(res_json["items"]) == 3

    def test_get_simple_item_query(self):
        response = self.client.get("/item/query/another")
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert len(res_json["results"]) == 1
        assert res_json["results"][0]["title"] == "another title"

    def test_query_item_string(self):
        response = self.client.get("/item/query/testattr%3aABC")  # testattr:ABC
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert len(res_json["results"]) == 1
        assert res_json["results"][0]["title"] == "and a third"

    def test_query_item_regex(self):
        response = self.client.get(
            "/item/query/testattr%3a%3a[A-C]%2b"
        )  # testattr::[A-C]+
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert len(res_json["results"]) == 1
        assert res_json["results"][0]["title"] == "and a third"

    def test_query_item_regex_backslash(self):
        response = self.client.get(
            "/item/query/testattr%3a%3a%5cw%2b"
        )  # testattr::\w+
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert len(res_json["results"]) == 1
        assert res_json["results"][0]["title"] == "and a third"

    def test_query_item_path(self):
        """Note: path queries are special: the query item must match the path
        from the root all the way to a directory, so this matches 1 item"""
        """ Note: filesystem separators in the query must be '\' """

        response = self.client.get(
            "/item/query/path:" + self.path_prefix + "\\somewhere\\a"
        )
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert len(res_json["results"]) == 1
        assert res_json["results"][0]["title"] == "another title"

    def test_get_all_albums(self):
        response = self.client.get("/album/")
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        response_albums = [album["album"] for album in res_json["albums"]]
        assert Counter(response_albums) == {"album": 1, "other album": 1}

    def test_get_single_album_by_id(self):
        response = self.client.get("/album/2")
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert res_json["id"] == 2
        assert res_json["album"] == "other album"

    def test_get_multiple_albums_by_id(self):
        response = self.client.get("/album/1,2")
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        response_albums = [album["album"] for album in res_json["albums"]]
        assert Counter(response_albums) == {"album": 1, "other album": 1}

    def test_get_album_empty_query(self):
        response = self.client.get("/album/query/")
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert len(res_json["albums"]) == 2

    def test_get_simple_album_query(self):
        response = self.client.get("/album/query/other")
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert len(res_json["results"]) == 1
        assert res_json["results"][0]["album"] == "other album"
        assert res_json["results"][0]["id"] == 2

    def test_get_album_details(self):
        response = self.client.get("/album/2?expand")
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert len(res_json["items"]) == 2
        assert res_json["items"][0]["album"] == "other album"
        assert res_json["items"][1]["album"] == "other album"
        response_track_titles = {item["title"] for item in res_json["items"]}
        assert response_track_titles == {"title", "and a third"}

    def test_query_album_string(self):
        response = self.client.get(
            "/album/query/albumtest%3axy"
        )  # albumtest:xy
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert len(res_json["results"]) == 1
        assert res_json["results"][0]["album"] == "album"

    def test_query_album_artpath_regex(self):
        response = self.client.get(
            "/album/query/artpath%3a%3aart_"
        )  # artpath::art_
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert len(res_json["results"]) == 1
        assert res_json["results"][0]["album"] == "other album"

    def test_query_album_regex_backslash(self):
        response = self.client.get(
            "/album/query/albumtest%3a%3a%5cw%2b"
        )  # albumtest::\w+
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert len(res_json["results"]) == 1
        assert res_json["results"][0]["album"] == "album"

    def test_get_stats(self):
        response = self.client.get("/stats")
        res_json = json.loads(response.data.decode("utf-8"))

        assert response.status_code == 200
        assert res_json["items"] == 3
        assert res_json["albums"] == 2

    def test_delete_item_id(self):
        web.app.config["READONLY"] = False

        # Create a temporary item
        item_id = self.lib.add(
            Item(title="test_delete_item_id", test_delete_item_id=1)
        )

        # Check we can find the temporary item we just created
        response = self.client.get(f"/item/{item_id}")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200
        assert res_json["id"] == item_id

        # Delete item by id
        response = self.client.delete(f"/item/{item_id}")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200

        # Check the item has gone
        response = self.client.get(f"/item/{item_id}")
        assert response.status_code == 404
        # Note: if this fails, the item may still be around
        # and may cause other tests to fail

    def test_delete_item_without_file(self):
        web.app.config["READONLY"] = False

        # Create an item with a file
        ipath = self.temp_path / "testfile1.mp3"
        shutil.copy(_common.RSRC / "full.mp3", ipath)
        assert ipath.exists()
        item_id = self.lib.add(Item.from_path(ipath))

        # Check we can find the temporary item we just created
        response = self.client.get(f"/item/{item_id}")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200
        assert res_json["id"] == item_id

        # Delete item by id, without deleting file
        response = self.client.delete(f"/item/{item_id}")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200

        # Check the item has gone
        response = self.client.get(f"/item/{item_id}")
        assert response.status_code == 404

        # Check the file has not gone
        assert ipath.exists()
        ipath.unlink()

    def test_delete_item_with_file(self):
        web.app.config["READONLY"] = False

        # Create an item with a file
        ipath = self.temp_path / "testfile2.mp3"
        shutil.copy(_common.RSRC / "full.mp3", ipath)
        assert ipath.exists()
        item_id = self.lib.add(Item.from_path(ipath))

        # Check we can find the temporary item we just created
        response = self.client.get(f"/item/{item_id}")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200
        assert res_json["id"] == item_id

        # Delete item by id, with file
        response = self.client.delete(f"/item/{item_id}?delete")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200

        # Check the item has gone
        response = self.client.get(f"/item/{item_id}")
        assert response.status_code == 404

        # Check the file has gone
        assert not ipath.exists()

    def test_delete_item_query(self):
        web.app.config["READONLY"] = False

        # Create a temporary item
        self.lib.add(
            Item(title="test_delete_item_query", test_delete_item_query=1)
        )

        # Check we can find the temporary item we just created
        response = self.client.get("/item/query/test_delete_item_query")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200
        assert len(res_json["results"]) == 1

        # Delete item by query
        response = self.client.delete("/item/query/test_delete_item_query")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200

        # Check the item has gone
        response = self.client.get("/item/query/test_delete_item_query")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200
        assert len(res_json["results"]) == 0

    def test_delete_item_all_fails(self):
        """DELETE is not supported for list all"""

        web.app.config["READONLY"] = False

        # Delete all items
        response = self.client.delete("/item/")
        assert response.status_code == 405

        # Note: if this fails, all items have gone and rest of
        # tests will fail!

    def test_delete_item_id_readonly(self):
        web.app.config["READONLY"] = True

        # Create a temporary item
        item_id = self.lib.add(
            Item(title="test_delete_item_id_ro", test_delete_item_id_ro=1)
        )

        # Check we can find the temporary item we just created
        response = self.client.get(f"/item/{item_id}")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200
        assert res_json["id"] == item_id

        # Try to delete item by id
        response = self.client.delete(f"/item/{item_id}")
        assert response.status_code == 405

        # Check the item has not gone
        response = self.client.get(f"/item/{item_id}")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200
        assert res_json["id"] == item_id

        # Remove it
        self.lib.get_item(item_id).remove()

    def test_delete_item_query_readonly(self):
        web.app.config["READONLY"] = True

        # Create a temporary item
        item_id = self.lib.add(
            Item(title="test_delete_item_q_ro", test_delete_item_q_ro=1)
        )

        # Check we can find the temporary item we just created
        response = self.client.get("/item/query/test_delete_item_q_ro")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200
        assert len(res_json["results"]) == 1

        # Try to delete item by query
        response = self.client.delete("/item/query/test_delete_item_q_ro")
        assert response.status_code == 405

        # Check the item has not gone
        response = self.client.get("/item/query/test_delete_item_q_ro")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200
        assert len(res_json["results"]) == 1

        # Remove it
        self.lib.get_item(item_id).remove()

    def test_delete_album_id(self):
        web.app.config["READONLY"] = False

        # Create a temporary album
        album_id = self.lib.add(
            Album(album="test_delete_album_id", test_delete_album_id=1)
        )

        # Check we can find the temporary album we just created
        response = self.client.get(f"/album/{album_id}")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200
        assert res_json["id"] == album_id

        # Delete album by id
        response = self.client.delete(f"/album/{album_id}")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200

        # Check the album has gone
        response = self.client.get(f"/album/{album_id}")
        assert response.status_code == 404
        # Note: if this fails, the album may still be around
        # and may cause other tests to fail

    def test_delete_album_query(self):
        web.app.config["READONLY"] = False

        # Create a temporary album
        self.lib.add(
            Album(album="test_delete_album_query", test_delete_album_query=1)
        )

        # Check we can find the temporary album we just created
        response = self.client.get("/album/query/test_delete_album_query")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200
        assert len(res_json["results"]) == 1

        # Delete album
        response = self.client.delete("/album/query/test_delete_album_query")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200

        # Check the album has gone
        response = self.client.get("/album/query/test_delete_album_query")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200
        assert len(res_json["results"]) == 0

    def test_delete_album_all_fails(self):
        """DELETE is not supported for list all"""

        web.app.config["READONLY"] = False

        # Delete all albums
        response = self.client.delete("/album/")
        assert response.status_code == 405

        # Note: if this fails, all albums have gone and rest of
        # tests will fail!

    def test_delete_album_id_readonly(self):
        web.app.config["READONLY"] = True

        # Create a temporary album
        album_id = self.lib.add(
            Album(album="test_delete_album_id_ro", test_delete_album_id_ro=1)
        )

        # Check we can find the temporary album we just created
        response = self.client.get(f"/album/{album_id}")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200
        assert res_json["id"] == album_id

        # Try to delete album by id
        response = self.client.delete(f"/album/{album_id}")
        assert response.status_code == 405

        # Check the item has not gone
        response = self.client.get(f"/album/{album_id}")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200
        assert res_json["id"] == album_id

        # Remove it
        self.lib.get_album(album_id).remove()

    def test_delete_album_query_readonly(self):
        web.app.config["READONLY"] = True

        # Create a temporary album
        album_id = self.lib.add(
            Album(
                album="test_delete_album_query_ro", test_delete_album_query_ro=1
            )
        )

        # Check we can find the temporary album we just created
        response = self.client.get("/album/query/test_delete_album_query_ro")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200
        assert len(res_json["results"]) == 1

        # Try to delete album
        response = self.client.delete("/album/query/test_delete_album_query_ro")
        assert response.status_code == 405

        # Check the album has not gone
        response = self.client.get("/album/query/test_delete_album_query_ro")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200
        assert len(res_json["results"]) == 1

        # Remove it
        self.lib.get_album(album_id).remove()

    def test_patch_item_id(self):
        # Note: PATCH is currently only implemented for track items, not albums

        web.app.config["READONLY"] = False

        # Create a temporary item
        item_id = self.lib.add(
            Item(
                title="test_patch_item_id", test_patch_f1=1, test_patch_f2="Old"
            )
        )

        # Check we can find the temporary item we just created
        response = self.client.get(f"/item/{item_id}")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200
        assert res_json["id"] == item_id
        assert res_json["test_patch_f1"] == "1"
        assert res_json["test_patch_f2"] == "Old"

        # Patch item by id
        # patch_json = json.JSONEncoder().encode({"test_patch_f2": "New"}]})
        response = self.client.patch(
            f"/item/{item_id}", json={"test_patch_f2": "New"}
        )
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200
        assert res_json["id"] == item_id
        assert res_json["test_patch_f1"] == "1"
        assert res_json["test_patch_f2"] == "New"

        # Check the update has really worked
        response = self.client.get(f"/item/{item_id}")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200
        assert res_json["id"] == item_id
        assert res_json["test_patch_f1"] == "1"
        assert res_json["test_patch_f2"] == "New"

        # Remove the item
        self.lib.get_item(item_id).remove()

    def test_patch_item_id_readonly(self):
        # Note: PATCH is currently only implemented for track items, not albums

        web.app.config["READONLY"] = True

        # Create a temporary item
        item_id = self.lib.add(
            Item(
                title="test_patch_item_id_ro",
                test_patch_f1=2,
                test_patch_f2="Old",
            )
        )

        # Check we can find the temporary item we just created
        response = self.client.get(f"/item/{item_id}")
        res_json = json.loads(response.data.decode("utf-8"))
        assert response.status_code == 200
        assert res_json["id"] == item_id
        assert res_json["test_patch_f1"] == "2"
        assert res_json["test_patch_f2"] == "Old"

        # Patch item by id
        # patch_json = json.JSONEncoder().encode({"test_patch_f2": "New"})
        response = self.client.patch(
            f"/item/{item_id}", json={"test_patch_f2": "New"}
        )
        assert response.status_code == 405

        # Remove the item
        self.lib.get_item(item_id).remove()

    def test_get_item_file(self):
        ipath = self.temp_path / "testfile2.mp3"
        shutil.copy(_common.RSRC / "full.mp3", ipath)
        assert ipath.exists()
        item_id = self.lib.add(Item.from_path(ipath))

        response = self.client.get(f"/item/{item_id}/file")

        assert response.status_code == 200


class TestWebXSS(WebPluginMixin, PytestTestHelper):
    """Tests for XSS vulnerability in the web plugin templates.

    These tests verify that the Underscore.js templates in index.html use
    the escaping syntax (<%- %) instead of the non-escaping syntax (<%= %).

    In Underscore.js 1.2.2 (used by beets):
    - <%= variable %> does NOT escape HTML (vulnerable to XSS)
    - <%- variable %> DOES escape HTML (safe)

    This was reported in
    https://github.com/beetbox/beets/security/advisories/GHSA-3gxm-wfjx-m847
    and remediated in
    https://github.com/beetbox/beets/commit/75f0d8f4899e61afb939adf02dcfb078aed23a6a
    """

    def test_templates_use_escaping_syntax(self):
        """Verify that all Underscore.js templates use <%- %> for escaping.

        This test requests the index.html page and checks that all
        user data interpolations in the Underscore.js templates use
        the escaping syntax (<%- %) rather than the non-escaping syntax (<%= %).
        """
        import re

        # Request the index.html page
        response = self.client.get("/")
        html = response.data.decode("utf-8")

        # Extract the template scripts from the HTML
        # The templates are in <script type="text/template"> blocks
        template_pattern = r'<script type="text/template"[^>]*>(.*?)</script>'
        templates = re.findall(template_pattern, html, re.DOTALL)

        # Combine all template content for checking
        all_template_content = "\n".join(templates)

        # Check that no <%= %> (non-escaping) tags exist for user data
        # We look for <%= followed by a variable name (word characters)
        non_escaping_pattern = r"<%=\s*(\w+)\s*%>"
        non_escaping_matches = re.findall(
            non_escaping_pattern, all_template_content
        )

        # List of fields that should be escaped (user-controlled data)
        user_data_fields = [
            "title",
            "artist",
            "album",
            "year",
            "track",
            "tracktotal",
            "disc",
            "disctotal",
            "length",
            "format",
            "bitrate",
            "mb_trackid",
            "id",
            "lyrics",
            "comments",
        ]

        # Check if any user data fields are using non-escaping <%= %>
        vulnerable_fields = [
            field for field in non_escaping_matches if field in user_data_fields
        ]

        # If we found any user data fields using <%= %>, the templates are vulnerable
        assert len(vulnerable_fields) == 0, (
            "Found non-escaping <%= %> tags for user data fields: "
            f"{vulnerable_fields}. "
            "These should use <%- %> for HTML escaping to prevent XSS."
        )

        # Also verify that escaping tags (<%- %>) are present for user data
        escaping_pattern = r"<%-\s*(\w+)\s*%>"
        escaping_matches = re.findall(escaping_pattern, all_template_content)

        # At least some user data fields should use escaping
        safe_fields = [
            field for field in escaping_matches if field in user_data_fields
        ]
        assert len(safe_fields) > 0, (
            "No escaping <%- %> tags found for user data fields. "
            "Templates should use <%- %> for HTML escaping."
        )



"""Tests for the 'ihate' plugin"""

from beets import importer
from beets.library import Item
from beetsplug.ihate import IHatePlugin


class TestIHatePlugin:
    def test_hate(self):
        match_pattern = {}
        test_item = Item(
            genres=["TestGenre"], album="TestAlbum", artist="TestArtist"
        )
        task = importer.SingletonImportTask(None, test_item)

        # Empty query should let it pass.
        assert not IHatePlugin.do_i_hate_this(task, match_pattern)

        # 1 query match.
        match_pattern = ["artist:bad_artist", "artist:TestArtist"]
        assert IHatePlugin.do_i_hate_this(task, match_pattern)

        # 2 query matches, either should trigger.
        match_pattern = ["album:test", "artist:testartist"]
        assert IHatePlugin.do_i_hate_this(task, match_pattern)

        # Query is blocked by AND clause.
        match_pattern = ["album:notthis genres:testgenre"]
        assert not IHatePlugin.do_i_hate_this(task, match_pattern)

        # Both queries are blocked by AND clause with unmatched condition.
        match_pattern = [
            "album:notthis genres:testgenre",
            "artist:testartist album:notthis",
        ]
        assert not IHatePlugin.do_i_hate_this(task, match_pattern)

        # Only one query should fire.
        match_pattern = [
            "album:testalbum genres:testgenre",
            "artist:testartist album:notthis",
        ]
        assert IHatePlugin.do_i_hate_this(task, match_pattern)



from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor as BaseThreadPoolExecutor
from threading import Event

import pytest

from beets import metadata_plugins
from beets.test.helper import PluginMixin


class ErrorMetadataMockPlugin(metadata_plugins.MetadataSourcePlugin):
    """A metadata source plugin that raises errors in all its methods."""

    def candidates(self, *args, **kwargs):
        raise ValueError("Mocked error")

    def item_candidates(self, *args, **kwargs):
        for i in range(3):
            raise ValueError("Mocked error")
            yield  # This is just to make this a generator

    def album_for_id(self, *args, **kwargs):
        raise ValueError("Mocked error")

    def track_for_id(self, *args, **kwargs):
        raise ValueError("Mocked error")


class TestMetadataPluginsException(PluginMixin):
    """Check that errors during the metadata plugins do not crash beets.
    They should be logged as errors instead.
    """

    @pytest.fixture(autouse=True)
    def setup(self):
        metadata_plugins.find_metadata_source_plugins.cache_clear()
        metadata_plugins.get_metadata_source.cache_clear()
        self.register_plugin(ErrorMetadataMockPlugin)
        yield
        self.unload_plugins()

    @pytest.fixture
    def call_method(self, method_name, args):
        def _call():
            result = getattr(metadata_plugins, method_name)(*args)
            return list(result) if isinstance(result, Iterable) else result

        return _call

    @pytest.mark.parametrize(
        "method_name,args",
        [
            ("candidates", ()),
            ("item_candidates", ()),
            ("albums_for_ids", (["some_id"],)),
            ("tracks_for_ids", (["some_id"],)),
            ("album_for_id", ("some_id", "ErrorMetadataMock")),
            ("track_for_id", ("some_id", "ErrorMetadataMock")),
        ],
    )
    def test_logging(self, caplog, call_method, method_name):
        self.config["raise_on_error"] = False

        call_method()

        assert (
            f"Error in 'ErrorMetadataMock.{method_name}': Mocked error"
            in caplog.text
        )

    @pytest.mark.parametrize(
        "method_name,args",
        [
            ("candidates", ()),
            ("item_candidates", ()),
            ("albums_for_ids", (["some_id"],)),
            ("tracks_for_ids", (["some_id"],)),
            ("album_for_id", ("some_id", "ErrorMetadataMock")),
            ("track_for_id", ("some_id", "ErrorMetadataMock")),
        ],
    )
    def test_raising(self, call_method):
        self.config["raise_on_error"] = True

        with pytest.raises(ValueError, match="Mocked error"):
            call_method()


class TestSearchApiMetadataSourcePlugin(PluginMixin):
    plugin = "none"
    preload_plugin = False

    class RaisingSearchApiMetadataMockPlugin(
        metadata_plugins.SearchApiMetadataSourcePlugin[
            metadata_plugins.IDResponse
        ]
    ):
        def get_search_query_with_filters(self, _):
            return "", {}

        def get_search_response(self, _):
            raise ValueError("Search failure")

        def album_for_id(self, _):
            return None

        def track_for_id(self, _):
            return None

    @pytest.fixture
    def search_plugin(self):
        return self.RaisingSearchApiMetadataMockPlugin()

    def test_search_api_returns_empty_when_raise_on_error_disabled(
        self, config, search_plugin, caplog
    ):
        config["raise_on_error"] = False

        assert search_plugin._search_api("track", "query", {}) == ()
        assert "Search failure" in caplog.text

    def test_search_api_raises_when_raise_on_error_enabled(
        self, config, search_plugin
    ):
        config["raise_on_error"] = True

        with pytest.raises(ValueError, match="Search failure"):
            search_plugin._search_api("track", "query", {})


def test_albums_for_ids_calls_each_plugin_once(monkeypatch):
    start_workers = Event()

    class Plugin:
        def __init__(self, name):
            self.data_source = name
            self.calls: list[list[str]] = []

        def albums_for_ids(self, ids):
            self.calls.append(list(ids))
            return [self.data_source]

    plugins = [Plugin("discogs"), Plugin("musicbrainz"), Plugin("tidal")]

    class PluginSequence:
        def __iter__(self):
            yield from plugins
            start_workers.set()

    class DelayedStartExecutor:
        def __init__(self):
            self._executor = BaseThreadPoolExecutor(max_workers=1)

        def submit(self, fn, *args, **kwargs):
            return self._executor.submit(self._run, fn, args, kwargs)

        def _run(self, fn, args, kwargs):
            start_workers.wait()
            return fn(*args, **kwargs)

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc, tb):
            self._executor.shutdown(wait=True)
            return False

    monkeypatch.setattr(
        metadata_plugins,
        "find_metadata_source_plugins",
        lambda: PluginSequence(),
    )
    monkeypatch.setattr(
        metadata_plugins, "ThreadPoolExecutor", DelayedStartExecutor
    )

    assert set(metadata_plugins.albums_for_ids(["42"])) == {
        "discogs",
        "musicbrainz",
        "tidal",
    }
    assert [plugin.calls for plugin in plugins] == [
        [["42"]],
        [["42"]],
        [["42"]],
    ]



import os
import shutil

from beets import library
from beets.test.helper import BeetsTestCase
from beets.ui.commands.move import move_items
from beets.util import syspath


class MoveTest(BeetsTestCase):
    def setUp(self):
        super().setUp()

        self.initial_item_path = self.lib_path / "srcfile"
        shutil.copy(self.resource_path, self.initial_item_path)

        # Add a file to the library but don't copy it in yet.
        self.i = library.Item.from_path(self.initial_item_path)
        self.lib.add(self.i)
        self.album = self.lib.add_album([self.i])

        # Alternate destination directory.
        self.otherdir = self.temp_path / "testotherdir"

    def _move(
        self,
        query=(),
        dest=None,
        copy=False,
        album=False,
        pretend=False,
        export=False,
    ):
        move_items(self.lib, dest, query, copy, album, pretend, export=export)

    def test_move_item(self):
        self._move()
        self.i.load()
        assert b"libdir" in self.i.path
        assert self.i.filepath.exists()
        assert not self.initial_item_path.exists()

    def test_copy_item(self):
        self._move(copy=True)
        self.i.load()
        assert b"libdir" in self.i.path
        assert self.i.filepath.exists()
        assert self.initial_item_path.exists()

    def test_move_album(self):
        self._move(album=True)
        self.i.load()
        assert b"libdir" in self.i.path
        assert self.i.filepath.exists()
        assert not self.initial_item_path.exists()

    def test_copy_album(self):
        self._move(copy=True, album=True)
        self.i.load()
        assert b"libdir" in self.i.path
        assert self.i.filepath.exists()
        assert self.initial_item_path.exists()

    def test_move_item_custom_dir(self):
        self._move(dest=self.otherdir)
        self.i.load()
        assert b"testotherdir" in self.i.path
        assert self.i.filepath.exists()
        assert not self.initial_item_path.exists()

    def test_move_album_custom_dir(self):
        self._move(dest=self.otherdir, album=True)
        self.i.load()
        assert b"testotherdir" in self.i.path
        assert self.i.filepath.exists()
        assert not self.initial_item_path.exists()

    def test_pretend_move_item(self):
        self._move(dest=self.otherdir, pretend=True)
        self.i.load()
        assert self.i.filepath == self.initial_item_path

    def test_pretend_move_album(self):
        self._move(album=True, pretend=True)
        self.i.load()
        assert self.i.filepath == self.initial_item_path

    def test_export_item_custom_dir(self):
        self._move(dest=self.otherdir, export=True)
        self.i.load()
        assert self.i.filepath == self.initial_item_path
        assert self.otherdir.exists()

    def test_export_album_custom_dir(self):
        self._move(dest=self.otherdir, album=True, export=True)
        self.i.load()
        assert self.i.filepath == self.initial_item_path
        assert self.otherdir.exists()

    def test_pretend_export_item(self):
        self._move(dest=self.otherdir, pretend=True, export=True)
        self.i.load()
        assert self.i.filepath == self.initial_item_path
        assert not self.otherdir.exists()

    def test_move_missing_singleton_continues(self):
        self.i.load()
        old_path = self.i.path
        os.remove(syspath(old_path))
        self._move()
        self.i.load()
        assert self.i.path == old_path

    def test_move_album_with_missing_track(self):
        self.i.load()
        old_i_path = self.i.path
        os.remove(syspath(old_i_path))

        i2_path = self.lib_path / "srcfile2"
        shutil.copy(self.resource_path, i2_path)
        i2 = library.Item.from_path(i2_path)
        self.lib.add(i2)
        i2.album_id = self.album.id
        i2.store()

        self._move(album=True)
        self.i.load()
        i2.load()
        assert self.i.path == old_i_path
        assert i2.filepath.is_relative_to(self.lib_path)
        assert i2.filepath.exists()

    def test_move_item_skips_missing_file(self):
        self.i.load()
        old_path = self.i.path
        os.remove(syspath(old_path))
        self.i.move()
        assert self.i.path == old_path



from __future__ import annotations

import fnmatch
import os
import shlex
import sys
from typing import TYPE_CHECKING

import pytest
from mediafile import MediaFile

from beets import util
from beets.library import Item
from beets.test import _common
from beets.test.helper import (
    AsIsImporterMixin,
    ImportHelper,
    IOMixin,
    PluginTestHelper,
)
from beetsplug import convert

if TYPE_CHECKING:
    from pathlib import Path

_p = pytest.param


class ConvertPluginHelper(IOMixin, PluginTestHelper):
    db_on_disk = True
    plugin = "convert"

    def setup_beets(self):
        super().setup_beets()
        self.convert_dest = self.temp_path / "convert_dest"
        self.config["convert"] = {"dest": str(self.convert_dest)}

    def tagged_copy_cmd(self, tag):
        """Return a conversion command that copies files and appends
        `tag` to the copy.
        """
        # A Python script that copies the file and appends a tag.
        stub = str(_common.RSRC / "convert_stub.py")
        return f"{shlex.quote(sys.executable)} {shlex.quote(stub)} $source $dest {tag}"

    def file_endswith(self, path: Path, tag: str):
        """Check the path is a file and if its content ends with `tag`."""
        assert path.exists()
        assert path.is_file()
        return path.read_bytes().endswith(tag.encode("utf-8"))


class TestImportConvert(AsIsImporterMixin, ImportHelper, ConvertPluginHelper):
    def setup_beets(self):
        super().setup_beets()
        self.config["convert"] = {
            "command": self.tagged_copy_cmd("convert"),
            # Enforce running convert
            "max_bitrate": 1,
            "auto": True,
            "quiet": False,
        }

    def test_import_converted(self):
        self.run_asis_importer()
        item = self.lib.items().get()
        assert self.file_endswith(item.filepath, "convert")

    # FIXME: fails on windows
    @pytest.mark.skipif(sys.platform == "win32", reason="win32")
    def test_import_original_on_convert_error(self):
        # `false` exits with non-zero code
        self.config["convert"]["command"] = "false"
        self.run_asis_importer()

        item = self.lib.items().get()
        assert item is not None
        assert item.filepath.is_file()

    def test_delete_originals(self):
        self.config["convert"]["delete_originals"] = True
        self.run_asis_importer()
        for path in self.importer.paths:
            for root, dirnames, filenames in os.walk(path):
                assert len(fnmatch.filter(filenames, "*.mp3")) == 0, (
                    f"Non-empty import directory {util.displayable_path(path)}"
                )


class ConvertCommand:
    """A mixin providing a utility method to run the `convert`command
    in tests.
    """

    def run_convert_path(self, item, *args):
        """Run the `convert` command on a given path."""
        return self.run_command("convert", *args, f"path:{item.filepath}")

    def run_convert(self, *args):
        """Run the `convert` command on `self.item`."""
        return self.run_convert_path(self.item, *args)


class TestConvertCli(ConvertPluginHelper, ConvertCommand):
    def setup_beets(self):
        super().setup_beets()
        self.album = self.add_album_fixture(ext="ogg")
        self.item = self.album.items()[0]

        self.converted_mp3 = self.convert_dest / "converted.mp3"
        self.config["convert"] = {
            "paths": {"default": "converted"},
            "format": "mp3",
            "formats": {
                "mp3": self.tagged_copy_cmd("mp3"),
                "ogg": self.tagged_copy_cmd("ogg"),
                "opus": {
                    "command": self.tagged_copy_cmd("opus"),
                    "extension": "ops",
                },
            },
        }

    def test_convert(self):
        self.io.addinput("y")
        self.run_convert()
        assert self.file_endswith(self.converted_mp3, "mp3")

    def test_convert_with_auto_confirmation(self):
        self.run_convert("--yes")
        assert self.file_endswith(self.converted_mp3, "mp3")

    def test_reject_confirmation(self):
        self.io.addinput("n")
        self.run_convert()
        assert not self.converted_mp3.exists()

    def test_convert_keep_new(self):
        assert self.item.filepath.suffix == ".ogg"

        self.io.addinput("y")
        self.run_convert("--keep-new")

        self.item.load()
        assert self.item.filepath.suffix == ".mp3"

    def test_format_option(self):
        self.io.addinput("y")
        self.run_convert("--format", "opus")
        assert self.file_endswith(self.convert_dest / "converted.ops", "opus")

    def test_embed_album_art(self):
        self.config["convert"]["embed"] = True
        image_path = _common.RSRC / "image-2x3.jpg"
        self.album.artpath = image_path
        self.album.store()
        image_data = image_path.read_bytes()

        self.io.addinput("y")
        self.run_convert()
        mediafile = MediaFile(self.converted_mp3)
        assert mediafile.images[0].data == image_data

    def test_copy_album_art_missing_source(self, caplog):
        # A missing/stale art source should be skipped instead of crashing
        # the conversion (see #4692).
        self.config["convert"]["copy_album_art"] = True
        self.album.artpath = _common.RSRC / "nonexistent.jpg"
        self.album.store()

        with caplog.at_level("INFO", logger="beets.convert"):
            self.run_command("convert", "-a", "--yes")

        assert any(
            "source file not found" in message for message in caplog.messages
        )
        assert self.file_endswith(self.converted_mp3, "mp3")

    def test_skip_existing(self):
        converted = self.converted_mp3
        self.touch(converted, content="XXX")
        self.run_convert("--yes")
        with open(converted) as f:
            assert f.read() == "XXX"

    def test_pretend(self):
        self.run_convert("--pretend")
        assert not self.converted_mp3.exists()

    def test_empty_query(self, caplog):
        with caplog.at_level("INFO", logger="beets.convert"):
            self.run_convert("An impossible query")
        assert caplog.messages[0] == "Empty query result."

    @pytest.mark.parametrize(
        "max_bitrate,convert_format,args,should_transcode",
        [
            _p(5000, "mp3", (), True, id="different-format-high-bitrate"),
            _p(5, "mp3", (), True, id="different-format-low-bitrate"),
            _p(None, "mp3", (), True, id="different-format-no-max-bitrate"),
            _p(5000, "ogg", (), False, id="same-format-high-bitrate"),
            _p(5000, "ogg", ("--force",), True, id="same-format-force"),
            _p(5, "ogg", (), True, id="same-format-low-bitrate"),
            _p(None, "ogg", (), False, id="same-format-no-max-bitrate"),
        ],
    )
    def test_transcode_selection(
        self, max_bitrate, convert_format, args, should_transcode
    ):
        if max_bitrate is not None:
            self.config["convert"]["max_bitrate"] = max_bitrate
        self.config["convert"]["format"] = convert_format

        self.io.addinput("y")
        self.run_convert(*args)

        converted_path = self.convert_dest / f"converted.{convert_format}"
        assert (
            self.file_endswith(converted_path, convert_format)
            is should_transcode
        )

    def test_playlist(self):
        self.io.addinput("y")
        self.run_convert("--playlist", "playlist.m3u8")
        assert (self.convert_dest / "playlist.m3u8").exists()

    def test_playlist_pretend(self):
        self.run_convert("--playlist", "playlist.m3u8", "--pretend")
        assert not (self.convert_dest / "playlist.m3u8").exists()

    @pytest.mark.parametrize(
        "config_overrides",
        [
            _p({"no_convert": "format:ogg"}, id="no-covert"),
            _p({"never_convert_lossy_files": True}, id="never-convert-lossy-files"),
        ],
    )  # fmt: skip
    def test_force_overrides(self, config_overrides):
        [item] = self.add_item_fixtures(ext="ogg")
        self.io.addinput("y")

        with self.configure_plugin(config_overrides):
            self.run_convert_path(item, "--format", "opus", "--force")

        converted = self.convert_dest / "converted.ops"
        assert self.file_endswith(converted, "opus")

    @pytest.mark.parametrize(
        "args,no_convert,expected_entry",
        [
            _p((), None, "converted.mp3", id="config-format"),
            _p(("--format", "opus"), None, "converted.ops", id="cli-format"),
            _p((), "format:ogg", "converted.ogg", id="not-transcoded"),
            _p(("--keep-new",), None, "converted.ogg", id="keep-new"),
        ],
    )
    def test_playlist_entry(self, args, no_convert, expected_entry):
        if no_convert:
            self.config["convert"]["no_convert"] = no_convert

        self.io.addinput("y")
        self.run_convert(*args, "--playlist", "playlist.m3u8")
        lines = (self.convert_dest / "playlist.m3u8").read_text().splitlines()
        assert lines[0] == "#EXTM3U"
        assert lines[1] == expected_entry


class TestNeverConvertLossyFiles(ConvertPluginHelper, ConvertCommand):
    """Test the effect of the `never_convert_lossy_files` option."""

    @pytest.mark.parametrize(
        "source_ext,never_convert_lossy_files,expected_ext,should_convert",
        [
            _p("flac", True, "mp3", True, id="lossless-converts-flag-on"),
            _p("flac", False, "mp3", True, id="lossless-converts-flag-off"),
            _p("ogg", False, "mp3", True, id="lossy-converts-allowed"),
            _p("ogg", True, "ogg", False, id="lossy-kept-prevented"),
        ],
    )
    def test_transcode(
        self,
        source_ext,
        never_convert_lossy_files,
        expected_ext,
        should_convert,
    ):
        [item] = self.add_item_fixtures(ext=source_ext)
        self.io.addinput("y")

        convert_fmt = "mp3"
        config = {
            "paths": {"default": "converted"},
            "format": convert_fmt,
            "formats": {convert_fmt: self.tagged_copy_cmd(convert_fmt)},
            "never_convert_lossy_files": never_convert_lossy_files,
        }
        with self.configure_plugin(config):
            self.run_convert_path(item)

        converted = self.convert_dest / f"converted.{expected_ext}"
        assert self.file_endswith(converted, convert_fmt) is should_convert


class TestNoConvert(PluginTestHelper):
    """Test the effect of the `no_convert` option."""

    plugin = "convert"

    @pytest.mark.parametrize(
        "config_value, should_skip",
        [
            ("", False),
            ("bitrate:320", False),
            ("bitrate:320 format:ogg", False),
            ("bitrate:320 , format:ogg", True),
        ],
    )
    def test_no_convert_skip(self, config, config_value, should_skip):
        item = Item(format="ogg", bitrate=256)
        config["convert"]["no_convert"] = config_value
        assert convert.ConvertPlugin().in_no_convert(item) == should_skip



import ctypes
import os
import sys
from typing import Any, ClassVar
from unittest import mock

from beets import importer, util
from beets.test.helper import (
    AutotagImportHelper,
    IOMixin,
    PluginMixin,
    PluginTestHelper,
)
from beetsplug.fetchart import CoverArtArchive, FetchArtPlugin, FileSystem


class TestFetchartImport(PluginMixin, AutotagImportHelper):
    plugin = "fetchart"
    preload_plugin = False
    plugin_defaults: ClassVar[Any] = {"sources": ["filesystem", "coverart"]}

    def setup_beets(self):
        super().setup_beets()

        self.prepare_album_for_import(1)
        self.setup_importer()

        self.local_art_patcher = mock.patch.object(
            FileSystem, "get", autospec=True, return_value=iter(())
        )
        self.local_art_mock = self.local_art_patcher.start()
        self.remote_art_patcher = mock.patch.object(
            CoverArtArchive, "get", autospec=True, return_value=iter(())
        )
        self.remote_art_mock = self.remote_art_patcher.start()

    def teardown_beets(self):
        super().teardown_beets()
        self.local_art_patcher.stop()
        self.remote_art_patcher.stop()

    def test_asis_skips_network_sources(self):
        self.importer.add_choice(importer.Action.ASIS)
        with self.configure_plugin(self.plugin_defaults):
            self.importer.run()
        self.local_art_mock.assert_called()
        self.remote_art_mock.assert_not_called()

    def test_apply_uses_network_sources(self):
        self.importer.add_choice(importer.Action.APPLY)
        with self.configure_plugin(self.plugin_defaults):
            self.importer.run()
        self.local_art_mock.assert_called()
        self.remote_art_mock.assert_called()

    def test_fetch_for_asis_uses_network_sources(self):
        self.importer.add_choice(importer.Action.ASIS)
        with self.configure_plugin(
            self.plugin_defaults | {"fetch_for_asis": True}
        ):
            self.importer.run()
        self.local_art_mock.assert_called()
        self.remote_art_mock.assert_called()


class TestFetchartCli(IOMixin, PluginTestHelper):
    plugin = "fetchart"

    def setup_beets(self):
        super().setup_beets()
        self.config["fetchart"]["cover_names"] = "c\xc3\xb6ver.jpg"
        self.config["art_filename"] = "mycover"
        self.album = self.add_album()

    def cover_path(self, ext: str = "jpg"):
        return os.path.join(self.album.path, f"mycover.{ext}".encode())

    def check_cover_is_stored(self, ext: str = "jpg"):
        assert self.album["artpath"] == self.cover_path(ext)
        with open(util.syspath(self.cover_path(ext))) as f:
            assert f.read() == "IMAGE"

    def hide_file_windows(self, path: bytes) -> None:
        if sys.platform == "win32":
            hidden_mask = 2
            assert ctypes.windll.kernel32.SetFileAttributesW(
                util.syspath(path), hidden_mask
            )

    def test_set_art_from_folder(self):
        self.touch(b"c\xc3\xb6ver.jpg", dir_=self.album.path, content="IMAGE")

        self.run_command("fetchart")

        self.album.load()
        self.check_cover_is_stored()

    def test_filesystem_does_not_pick_up_folder(self):
        os.makedirs(os.path.join(self.album.path, b"mycover.jpg"))
        self.run_command("fetchart")
        self.album.load()
        assert self.album["artpath"] is None

    def test_filesystem_does_not_pick_up_ignored_file(self):
        self.touch(b"co_ver.jpg", dir_=self.album.path, content="IMAGE")
        self.config["ignore"] = ["*_*"]
        self.run_command("fetchart")
        self.album.load()
        assert self.album["artpath"] is None

    def test_filesystem_picks_up_non_ignored_file(self):
        self.touch(b"cover.jpg", dir_=self.album.path, content="IMAGE")
        self.config["ignore"] = ["*_*"]
        self.run_command("fetchart")
        self.album.load()
        self.check_cover_is_stored()

    def test_filesystem_does_not_pick_up_hidden_file(self):
        path = self.touch(b".cover.jpg", dir_=self.album.path, content="IMAGE")
        self.hide_file_windows(path)
        self.config["ignore"] = []  # By default, ignore includes '.*'.
        self.config["ignore_hidden"] = True
        self.run_command("fetchart")
        self.album.load()
        assert self.album["artpath"] is None

    def test_filesystem_picks_up_non_hidden_file(self):
        self.touch(b"cover.jpg", dir_=self.album.path, content="IMAGE")
        self.config["ignore_hidden"] = True
        self.run_command("fetchart")
        self.album.load()
        self.check_cover_is_stored()

    def test_filesystem_picks_up_hidden_file(self):
        path = self.touch(b".cover.jpg", dir_=self.album.path, content="IMAGE")
        self.hide_file_windows(path)
        self.config["ignore"] = []  # By default, ignore includes '.*'.
        self.config["ignore_hidden"] = False
        self.run_command("fetchart")
        self.album.load()
        self.check_cover_is_stored()

    def test_filesystem_picks_up_webp_file(self):
        self.touch(b"cover.webp", dir_=self.album.path, content="IMAGE")
        self.run_command("fetchart")
        self.album.load()
        self.check_cover_is_stored("webp")

    def test_filesystem_picks_up_png_file(self):
        self.touch(b"cover.png", dir_=self.album.path, content="IMAGE")
        self.run_command("fetchart")
        self.album.load()
        self.check_cover_is_stored("png")

    def test_colorization(self):
        self.config["ui"]["color"] = True
        out = self.run_with_output("fetchart")
        assert " - the älbum: \x1b[1;31mno art found\x1b[39;49;00m\n" == out

    def test_set_art_oserror_is_handled_gracefully(self):
        """OSError (e.g. PermissionError) in set_art is logged as a warning,
        not an unhandled crash. Regression test for #6193.
        """
        self.touch(b"c\xc3\xb6ver.jpg", dir_=self.album.path, content="IMAGE")
        with mock.patch(
            "beets.library.Album.set_art",
            side_effect=PermissionError("[WinError 32] file in use"),
        ):
            out = self.run_with_output("fetchart")
        self.album.load()
        assert "error writing album art" in out
        assert self.album["artpath"] is None

    def test_sources_is_a_string(self):
        self.config["fetchart"].set({"sources": "filesystem"})
        fa = FetchArtPlugin()
        assert len(fa.sources) == 1
        assert isinstance(fa.sources[0], FileSystem)

    def test_sources_is_an_asterisk(self):
        self.config["fetchart"].set({"sources": "*"})
        fa = FetchArtPlugin()
        assert len(fa.sources) == 10

    def test_sources_is_a_string_list(self):
        self.config["fetchart"].set({"sources": ["filesystem", "coverart"]})
        fa = FetchArtPlugin()
        assert len(fa.sources) == 3

    def test_sources_is_a_mapping_list(self):
        self.config["fetchart"].set(
            {"sources": {"filesystem": "*", "coverart": "*"}}
        )
        fa = FetchArtPlugin()
        assert len(fa.sources) == 3



"""Send the results of a query to the configured music player as a playlist."""

import random
import shlex
import subprocess
from os.path import relpath

from beets import config, ui, util
from beets.exceptions import UserError
from beets.plugins import BeetsPlugin
from beets.ui import Subcommand
from beets.util import PromptChoice, get_temp_filename
from beets.util.color import colorize

# Indicate where arguments should be inserted into the command string.
# If this is missing, they're placed at the end.
ARGS_MARKER = "$args"

# Indicate where the playlist file (with absolute path) should be inserted into
# the command string. If this is missing, its placed at the end, but before
# arguments.
PLS_MARKER = "$playlist"


def play(
    command_str,
    selection,
    paths,
    open_args,
    log,
    item_type="track",
    keep_open=False,
):
    """Play items in paths with command_str and optional arguments. If
    keep_open, return to beets, otherwise exit once command runs.
    """
    # Print number of tracks or albums to be played, log command to be run.
    item_type += "s" if len(selection) > 1 else ""
    ui.print_(f"Playing {len(selection)} {item_type}.")
    log.debug("executing command: {} {!r}", command_str, open_args)

    try:
        if keep_open:
            command = shlex.split(command_str)
            command = command + open_args
            subprocess.call(command)
        else:
            util.interactive_open(open_args, command_str)
    except OSError as exc:
        raise UserError(f"Could not play the query: {exc}")


class PlayPlugin(BeetsPlugin):
    def __init__(self):
        super().__init__()

        config["play"].add(
            {
                "command": None,
                "use_folders": False,
                "relative_to": None,
                "raw": False,
                "warning_threshold": 100,
                "bom": False,
            }
        )

        self.register_listener(
            "before_choose_candidate", self.before_choose_candidate_listener
        )

    def commands(self):
        play_command = Subcommand(
            "play", help="send music to a player as a playlist"
        )
        play_command.parser.add_album_option()
        play_command.parser.add_option(
            "-A",
            "--args",
            action="store",
            help="add additional arguments to the command",
        )
        play_command.parser.add_option(
            "-R",
            "--randomize",
            action="store_true",
            help="randomize the order of playlist entries",
        )
        play_command.parser.add_option(
            "-y",
            "--yes",
            action="store_true",
            help="skip the warning threshold",
        )
        play_command.func = self._play_command
        return [play_command]

    def _play_command(self, lib, opts, args):
        """The CLI command function for `beet play`. Create a list of paths
        from query, determine if tracks or albums are to be played.
        """
        use_folders = config["play"]["use_folders"].get(bool)
        relative_to = config["play"]["relative_to"].get()
        if relative_to:
            relative_to = util.normpath(relative_to)
        # Perform search by album and add folders rather than tracks to
        # playlist.
        if opts.album:
            selection = lib.albums(args)
            paths = []

            sort = lib.get_default_album_sort()
            for album in selection:
                if use_folders:
                    paths.append(album.item_dir())
                else:
                    paths.extend(item.path for item in sort.sort(album.items()))
            item_type = "album"

        # Perform item query and add tracks to playlist.
        else:
            selection = lib.items(args)
            paths = [item.path for item in selection]
            item_type = "track"

        if relative_to:
            paths = [relpath(path, relative_to) for path in paths]

        if not selection:
            ui.print_(colorize("text_warning", f"No {item_type} to play."))
            return

        if opts.randomize:
            random.shuffle(paths)

        open_args = self._playlist_or_paths(paths)
        open_args_str = [
            p.decode("utf-8") for p in self._playlist_or_paths(paths)
        ]
        command_str = self._command_str(opts.args)

        if PLS_MARKER in command_str:
            if not config["play"]["raw"]:
                command_str = command_str.replace(
                    PLS_MARKER, "".join(open_args_str)
                )
                self._log.debug(
                    "command altered by PLS_MARKER to: {}", command_str
                )
                open_args = []
            else:
                command_str = command_str.replace(PLS_MARKER, " ")

        # Check if the selection exceeds configured threshold. If True,
        # cancel, otherwise proceed with play command.
        if opts.yes or not self._exceeds_threshold(
            selection, command_str, open_args, item_type
        ):
            play(command_str, selection, paths, open_args, self._log, item_type)

    def _command_str(self, args=None):
        """Create a command string from the config command and optional args."""
        command_str = config["play"]["command"].get()
        if not command_str:
            return util.open_anything()
        # Add optional arguments to the player command.
        if args:
            if ARGS_MARKER in command_str:
                return command_str.replace(ARGS_MARKER, args)
            return f"{command_str} {args}"
        # Don't include the marker in the command.
        return command_str.replace(f" {ARGS_MARKER}", "")

    def _playlist_or_paths(self, paths):
        """Return either the raw paths of items or a playlist of the items."""
        if config["play"]["raw"]:
            return paths
        return [self._create_tmp_playlist(paths)]

    def _exceeds_threshold(
        self, selection, command_str, open_args, item_type="track"
    ):
        """Prompt user whether to abort if playlist exceeds threshold. If
        True, cancel playback. If False, execute play command.
        """
        warning_threshold = config["play"]["warning_threshold"].get(int)

        # Warn user before playing any huge playlists.
        if warning_threshold and len(selection) > warning_threshold:
            if len(selection) > 1:
                item_type += "s"

            ui.print_(
                colorize(
                    "text_warning",
                    f"You are about to queue {len(selection)} {item_type}.",
                )
            )

            if ui.input_options(("Continue", "Abort")) == "a":
                return True

        return False

    def _create_tmp_playlist(self, paths_list):
        """Create a temporary .m3u file. Return the filename."""
        utf8_bom = config["play"]["bom"].get(bool)
        filename = get_temp_filename(__name__, suffix=".m3u")
        with open(filename, "wb") as m3u:
            if utf8_bom:
                m3u.write(b"\xef\xbb\xbf")

            for item in paths_list:
                m3u.write(item + b"\n")

        return filename

    def before_choose_candidate_listener(self, session, task):
        """Append a "Play" choice to the interactive importer prompt."""
        return [PromptChoice("y", "plaY", self.importer_play)]

    def importer_play(self, session, task):
        """Get items from current import task and send to play function."""
        selection = task.items
        paths = [item.path for item in selection]

        open_args = self._playlist_or_paths(paths)
        command_str = self._command_str()

        if not self._exceeds_threshold(selection, command_str, open_args):
            play(
                command_str,
                selection,
                paths,
                open_args,
                self._log,
                keep_open=True,
            )



import os
import unittest
from unittest.mock import Mock, patch

import pytest

from beets import config, library
from beets.autotag import AlbumInfo, AlbumMatch, TrackInfo, distance
from beets.exceptions import UserError
from beets.test import _common
from beets.test.helper import BeetsTestCase, IOMixin
from beets.ui.commands.import_ import import_files, paths_from_logfile
from beets.ui.commands.import_.display import show_change
from beets.ui.commands.import_.session import summarize_items


class ImportTest(BeetsTestCase):
    def test_quiet_timid_disallowed(self):
        config["import"]["quiet"] = True
        config["import"]["timid"] = True
        with pytest.raises(UserError):
            import_files(None, [], None)

    def test_parse_paths_from_logfile(self):
        if os.path.__name__ == "ntpath":
            logfile_content = (
                "import started Wed Jun 15 23:08:26 2022\n"
                "asis C:\\music\\Beatles, The\\The Beatles; C:\\music\\Beatles, The\\The Beatles\\CD 01; C:\\music\\Beatles, The\\The Beatles\\CD 02\n"  # noqa: E501
                "duplicate-replace C:\\music\\Bill Evans\\Trio '65\n"
                "skip C:\\music\\Michael Jackson\\Bad\n"
                "skip C:\\music\\Soulwax\\Any Minute Now\n"
            )
            expected_paths = [
                "C:\\music\\Beatles, The\\The Beatles",
                "C:\\music\\Michael Jackson\\Bad",
                "C:\\music\\Soulwax\\Any Minute Now",
            ]
        else:
            logfile_content = (
                "import started Wed Jun 15 23:08:26 2022\n"
                "asis /music/Beatles, The/The Beatles; /music/Beatles, The/The Beatles/CD 01; /music/Beatles, The/The Beatles/CD 02\n"  # noqa: E501
                "duplicate-replace /music/Bill Evans/Trio '65\n"
                "skip /music/Michael Jackson/Bad\n"
                "skip /music/Soulwax/Any Minute Now\n"
            )
            expected_paths = [
                "/music/Beatles, The/The Beatles",
                "/music/Michael Jackson/Bad",
                "/music/Soulwax/Any Minute Now",
            ]

        logfile = self.temp_path / "logfile.log"
        logfile.write_text(logfile_content)
        actual_paths = list(paths_from_logfile(logfile))
        assert actual_paths == expected_paths


@patch("beets.ui.term_width", Mock(return_value=54))
class ShowChangeTestCase(IOMixin, BeetsTestCase):
    def _show_change(self):
        """Return an unicode string representing the changes"""
        long_name = f"a{' very' * 10} long name"
        items = [
            _common.item(track=1, title="first title"),
            _common.item(track=2, title="", path=b"/path/to/file.mp3"),
            _common.item(track=3, title="caf\xe9"),
            _common.item(track=4, title=f"title with {long_name}"),
        ]
        info = AlbumInfo(
            album="caf\xe9",
            album_id="album id",
            artist="the artist",
            artist_id="artist id",
            tracks=[
                TrackInfo(title="first title", index=1),
                TrackInfo(title="second title", index=2),
                TrackInfo(title="third title", index=3),
                TrackInfo(title="fourth title", index=4),
            ],
        )
        item_info_pairs = list(zip(items, info.tracks))
        self.config["ui"]["color"] = False
        self.config["import"]["detail"] = True
        change_dist = distance(items, info, item_info_pairs)
        change_dist._penalties = {"album": [0.1], "artist": [0.1]}
        show_change(
            f"another artist with {long_name}",
            "another album",
            AlbumMatch(change_dist, info, dict(item_info_pairs)),
        )
        return self.io.getoutput()

    def test_newline_layout(self):
        self.config["ui"]["import"]["layout"] = "newline"
        msg = self._show_change()
        assert (
            msg
            == """
  Match (90.0%):
  the artist - café
  ≠ album, artist
  None, None, None, None, None, None, None
  ≠ Artist: another artist with a very very very very
    very very very very very very long name
   -> the artist
  ≠ Album: another album -> café
     * (#1) first title (1:00)
     ≠ (#2) file.mp3 (1:00)
      -> (#2) second title (0:00)
     ≠ (#3) café (1:00) -> (#3) third title (0:00)
     ≠ (#4) title with a very very very very very very
          very very very very long name (1:00)
      -> (#4) fourth title (0:00)
"""
        )

    def test_column_layout(self):
        self.config["ui"]["import"]["layout"] = "column"
        msg = self._show_change()
        assert (
            msg
            == """
  Match (90.0%):
  the artist - café
  ≠ album, artist
  None, None, None, None, None, None, None
  ≠ Artist: another artist -> the artist              
            with a very                               
            very very very                            
            very very very                            
            very very very                            
            long name                                 
  ≠ Album: another album -> café
     * (#1) first title (1:00)
     ≠ (#2) file.mp (1:00) -> (#2) second    (0:00)
            3                      title           
     ≠ (#3) café (1:00) -> (#3) third title (0:00)
     ≠ (#4) title   (1:00) -> (#4) fourth    (0:00)
            with a very            title           
            very very very                         
            very very very                         
            very very very                         
            long name                              
"""  # noqa: W291
        )


@patch("beets.library.Item.try_filesize", Mock(return_value=987))
class SummarizeItemsTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        item = library.Item()
        item.bitrate = 4321
        item.length = 10 * 60 + 54
        item.format = "F"
        self.item = item

    def test_summarize_item(self):
        summary = summarize_items([], True)
        assert summary == ""

        summary = summarize_items([self.item], True)
        assert summary == "F, 4kbps, 10:54, 987.0 B"

    def test_summarize_items(self):
        summary = summarize_items([], False)
        assert summary == "0 items"

        summary = summarize_items([self.item], False)
        assert summary == "1 items, F, 4kbps, 10:54, 987.0 B"

        # make a copy of self.item
        i2 = self.item.copy()

        summary = summarize_items([self.item, i2], False)
        assert summary == "2 items, F, 4kbps, 21:48, 1.9 KiB"

        i2.format = "G"
        summary = summarize_items([self.item, i2], False)
        assert summary == "2 items, F 1, G 1, 4kbps, 21:48, 1.9 KiB"

        summary = summarize_items([self.item, i2, i2], False)
        assert summary == "3 items, G 2, F 1, 4kbps, 32:42, 2.9 KiB"



from __future__ import annotations

import datetime
import re
from functools import cache, total_ordering
from typing import TYPE_CHECKING, Any

from jellyfish import levenshtein_distance
from unidecode import unidecode

from beets import config, metadata_plugins
from beets.util import as_string, cached_classproperty, get_most_common_tags
from beets.util.color import colorize

if TYPE_CHECKING:
    from collections.abc import Iterator, KeysView, Sequence

    from beets.library import Item
    from beets.util.color import ColorName

    from .hooks import AlbumInfo, TrackInfo

# Candidate distance scoring.

# Artist signals that indicate "various artists". These are used at the
# album level to determine whether a given release is likely a VA
# release and also on the track level to to remove the penalty for
# differing artists.
VA_ARTISTS = ("", "various artists", "various", "va", "unknown")

# Parameters for string distance function.
# Words that can be moved to the end of a string using a comma.
SD_END_WORDS = ["the", "a", "an"]
# Reduced weights for certain portions of the string.
SD_PATTERNS = [
    (r"^the ", 0.1),
    (r"[\[\(]?(ep|single)[\]\)]?", 0.0),
    (r"[\[\(]?(featuring|feat|ft)[\. :].+", 0.1),
    (r"\(.*?\)", 0.3),
    (r"\[.*?\]", 0.3),
    (r"(, )?(pt\.|part) .+", 0.2),
]
# Replacements to use before testing distance.
SD_REPLACE = [(r"&", "and")]


def _string_dist_basic(str1: str, str2: str) -> float:
    """Basic edit distance between two strings, ignoring
    non-alphanumeric characters and case. Comparisons are based on a
    transliteration/lowering to ASCII characters. Normalized by string
    length.
    """
    assert isinstance(str1, str)
    assert isinstance(str2, str)
    str1 = as_string(unidecode(str1))
    str2 = as_string(unidecode(str2))
    str1 = re.sub(r"[^a-z0-9]", "", str1.lower())
    str2 = re.sub(r"[^a-z0-9]", "", str2.lower())
    if not str1 and not str2:
        return 0.0
    return levenshtein_distance(str1, str2) / float(max(len(str1), len(str2)))


def string_dist(str1: str | None, str2: str | None) -> float:
    """Gives an "intuitive" edit distance between two strings. This is
    an edit distance, normalized by the string length, with a number of
    tweaks that reflect intuition about text.
    """
    if str1 is None and str2 is None:
        return 0.0
    if str1 is None or str2 is None:
        return 1.0

    str1 = str1.lower()
    str2 = str2.lower()

    # Don't penalize strings that move certain words to the end. For
    # example, "the something" should be considered equal to
    # "something, the".
    for word in SD_END_WORDS:
        if str1.endswith(f", {word}"):
            str1 = f"{word} {str1[: -len(word) - 2]}"
        if str2.endswith(f", {word}"):
            str2 = f"{word} {str2[: -len(word) - 2]}"

    # Perform a couple of basic normalizing substitutions.
    for pat, repl in SD_REPLACE:
        str1 = re.sub(pat, repl, str1)
        str2 = re.sub(pat, repl, str2)

    # Change the weight for certain string portions matched by a set
    # of regular expressions. We gradually change the strings and build
    # up penalties associated with parts of the string that were
    # deleted.
    base_dist = _string_dist_basic(str1, str2)
    penalty = 0.0
    for pat, weight in SD_PATTERNS:
        # Get strings that drop the pattern.
        case_str1 = re.sub(pat, "", str1)
        case_str2 = re.sub(pat, "", str2)

        if case_str1 != str1 or case_str2 != str2:
            # If the pattern was present (i.e., it is deleted in the
            # the current case), recalculate the distances for the
            # modified strings.
            case_dist = _string_dist_basic(case_str1, case_str2)
            case_delta = max(0.0, base_dist - case_dist)
            if case_delta == 0.0:
                continue

            # Shift our baseline strings down (to avoid rematching the
            # same part of the string) and add a scaled distance
            # amount to the penalties.
            str1 = case_str1
            str2 = case_str2
            base_dist = case_dist
            penalty += weight * case_delta

    return base_dist + penalty


@total_ordering
class Distance:
    """Keeps track of multiple distance penalties. Provides a single
    weighted distance for all penalties as well as a weighted distance
    for each individual penalty.
    """

    def __init__(self) -> None:
        self._penalties: dict[str, list[float]] = {}
        self.tracks: dict[TrackInfo, Distance] = {}

    @cached_classproperty
    def _weights(cls) -> dict[str, float]:
        """A dictionary from keys to floating-point weights."""
        weights_view = config["match"]["distance_weights"]
        weights = {}
        for key in weights_view.keys():
            weights[key] = weights_view[key].as_number()
        return weights

    @property
    def generic_penalty_keys(self) -> list[str]:
        return [
            k.replace("album_", "").replace("track_", "").replace("_", " ")
            for k in self._penalties
            if self[k]
        ]

    # Access the components and their aggregates.

    @property
    def distance(self) -> float:
        """Return a weighted and normalized distance across all
        penalties.
        """
        dist_max = self.max_distance
        if dist_max:
            return self.raw_distance / self.max_distance
        return 0.0

    @property
    def max_distance(self) -> float:
        """Return the maximum distance penalty (normalization factor)."""
        dist_max = 0.0
        for key, penalty in self._penalties.items():
            dist_max += len(penalty) * self._weights[key]
        return dist_max

    @property
    def raw_distance(self) -> float:
        """Return the raw (denormalized) distance."""
        dist_raw = 0.0
        for key, penalty in self._penalties.items():
            dist_raw += sum(penalty) * self._weights[key]
        return dist_raw

    @property
    def color(self) -> ColorName:
        if self.distance <= config["match"]["strong_rec_thresh"].as_number():
            return "text_success"
        if self.distance <= config["match"]["medium_rec_thresh"].as_number():
            return "text_warning"
        return "text_error"

    @property
    def string(self) -> str:
        return colorize(self.color, f"{(1 - self.distance) * 100:.1f}%")

    def items(self) -> list[tuple[str, float]]:
        """Return a list of (key, dist) pairs, with `dist` being the
        weighted distance, sorted from highest to lowest. Does not
        include penalties with a zero value.
        """
        list_ = []
        for key in self._penalties:
            dist = self[key]
            if dist:
                list_.append((key, dist))
        # Convert distance into a negative float we can sort items in
        # ascending order (for keys, when the penalty is equal) and
        # still get the items with the biggest distance first.
        return sorted(
            list_, key=lambda key_and_dist: (-key_and_dist[1], key_and_dist[0])
        )

    def __hash__(self) -> int:
        return id(self)

    def __eq__(self, other) -> bool:
        return self.distance == other

    # Behave like a float.

    def __lt__(self, other) -> bool:
        return self.distance < other

    def __float__(self) -> float:
        return self.distance

    def __sub__(self, other) -> float:
        return self.distance - other

    def __rsub__(self, other) -> float:
        return other - self.distance

    def __str__(self) -> str:
        return f"{self.distance:.2f}"

    # Behave like a dict.

    def __getitem__(self, key) -> float:
        """Returns the weighted distance for a named penalty."""
        dist = sum(self._penalties[key]) * self._weights[key]
        dist_max = self.max_distance
        if dist_max:
            return dist / dist_max
        return 0.0

    def __iter__(self) -> Iterator[tuple[str, float]]:
        return iter(self.items())

    def __len__(self) -> int:
        return len(self.items())

    def keys(self) -> KeysView[str]:
        return dict.fromkeys(key for key, _ in self.items()).keys()

    def update(self, dist: Distance):
        """Adds all the distance penalties from `dist`."""
        if not isinstance(dist, Distance):
            raise ValueError(
                f"`dist` must be a Distance object, not {type(dist)}"
            )
        for key, penalties in dist._penalties.items():
            self._penalties.setdefault(key, []).extend(penalties)

    # Adding components.

    def _eq(self, value1: re.Pattern[str] | Any, value2: Any) -> bool:
        """Returns True if `value1` is equal to `value2`. `value1` may
        be a compiled regular expression, in which case it will be
        matched against `value2`.
        """
        if isinstance(value1, re.Pattern):
            return bool(value1.match(value2))
        return value1 == value2

    def add(self, key: str, dist: float):
        """Adds a distance penalty. `key` must correspond with a
        configured weight setting. `dist` must be a float between 0.0
        and 1.0, and will be added to any existing distance penalties
        for the same key.
        """
        if not 0.0 <= dist <= 1.0:
            raise ValueError(f"`dist` must be between 0.0 and 1.0, not {dist}")
        self._penalties.setdefault(key, []).append(dist)

    def add_equality(
        self, key: str, value: Any, options: list[Any] | tuple[Any, ...] | Any
    ):
        """Adds a distance penalty of 1.0 if `value` doesn't match any
        of the values in `options`. If an option is a compiled regular
        expression, it will be considered equal if it matches against
        `value`.
        """
        if not isinstance(options, (list, tuple)):
            options = [options]
        for opt in options:
            if self._eq(opt, value):
                dist = 0.0
                break
        else:
            dist = 1.0
        self.add(key, dist)

    def add_expr(self, key: str, expr: bool):
        """Adds a distance penalty of 1.0 if `expr` evaluates to True,
        or 0.0.
        """
        if expr:
            self.add(key, 1.0)
        else:
            self.add(key, 0.0)

    def add_number(self, key: str, number1: int, number2: int):
        """Adds a distance penalty of 1.0 for each number of difference
        between `number1` and `number2`, or 0.0 when there is no
        difference. Use this when there is no upper limit on the
        difference between the two numbers.
        """
        diff = abs(number1 - number2)
        if diff:
            for i in range(diff):
                self.add(key, 1.0)
        else:
            self.add(key, 0.0)

    def add_priority(
        self, key: str, value: Any, options: list[Any] | tuple[Any, ...] | Any
    ):
        """Adds a distance penalty that corresponds to the position at
        which `value` appears in `options`. A distance penalty of 0.0
        for the first option, or 1.0 if there is no matching option. If
        an option is a compiled regular expression, it will be
        considered equal if it matches against `value`.
        """
        if not isinstance(options, (list, tuple)):
            options = [options]
        unit = 1.0 / (len(options) or 1)
        for i, opt in enumerate(options):
            if self._eq(opt, value):
                dist = i * unit
                break
        else:
            dist = 1.0
        self.add(key, dist)

    def add_ratio(self, key: str, number1: int | float, number2: int | float):
        """Adds a distance penalty for `number1` as a ratio of `number2`.
        `number1` is bound at 0 and `number2`.
        """
        number = float(max(min(number1, number2), 0))
        if number2:
            dist = number / number2
        else:
            dist = 0.0
        self.add(key, dist)

    def add_string(self, key: str, str1: str | None, str2: str | None):
        """Adds a distance penalty based on the edit distance between
        `str1` and `str2`.
        """
        dist = string_dist(str1, str2)
        self.add(key, dist)

    def add_data_source(self, before: str | None, after: str | None) -> None:
        if before != after and (
            before or len(metadata_plugins.find_metadata_source_plugins()) > 1
        ):
            self.add("data_source", metadata_plugins.get_penalty(after))


@cache
def get_track_length_grace() -> float:
    """Get cached grace period for track length matching."""
    return config["match"]["track_length_grace"].as_number()


@cache
def get_track_length_max() -> float:
    """Get cached maximum track length for track length matching."""
    return config["match"]["track_length_max"].as_number()


def track_index_changed(item: Item, track_info: TrackInfo) -> bool:
    """Returns True if the item and track info index is different. Tolerates
    per disc and per release numbering.
    """
    return item.track not in (track_info.medium_index, track_info.index)


def track_distance(
    item: Item, track_info: TrackInfo, incl_artist: bool = False
) -> Distance:
    """Determines the significance of a track metadata change. Returns a
    Distance object. `incl_artist` indicates that a distance component should
    be included for the track artist (i.e., for various-artist releases).

    ``track_length_grace`` and ``track_length_max`` configuration options are
    cached because this function is called many times during the matching
    process and their access comes with a performance overhead.
    """
    dist = Distance()

    # Length.
    if info_length := track_info.length:
        diff = abs(item.length - info_length) - get_track_length_grace()
        dist.add_ratio("track_length", diff, get_track_length_max())

    # Title.
    dist.add_string("track_title", item.title, track_info.title)

    # Artist. Only check if there is actually an artist in the track data.
    if (
        incl_artist
        and track_info.artist
        and item.artist.lower() not in VA_ARTISTS
    ):
        dist.add_string("track_artist", item.artist, track_info.artist)

    # Track index.
    if track_info.index and item.track:
        dist.add_expr("track_index", track_index_changed(item, track_info))

    # Track ID.
    if item.mb_trackid:
        dist.add_expr("track_id", item.mb_trackid != track_info.track_id)

    # Penalize mismatching disc numbers.
    if track_info.medium and item.disc:
        dist.add_expr("medium", item.disc != track_info.medium)

    dist.add_data_source(item.get("data_source"), track_info.data_source)

    return dist


def distance(
    items: Sequence[Item],
    album_info: AlbumInfo,
    item_info_pairs: list[tuple[Item, TrackInfo]],
) -> Distance:
    """Determines how "significant" an album metadata change would be.
    Returns a Distance object. `album_info` is an AlbumInfo object
    reflecting the album to be compared. `items` is a sequence of all
    Item objects that will be matched (order is not important).
    `mapping` is a dictionary mapping Items to TrackInfo objects; the
    keys are a subset of `items` and the values are a subset of
    `album_info.tracks`.
    """
    likelies, _ = get_most_common_tags(items)

    dist = Distance()

    # Artist, if not various.
    if not album_info.va:
        dist.add_string("artist", likelies["artist"], album_info.artist)

    # Album.
    dist.add_string("album", likelies["album"], album_info.album)

    preferred_config = config["match"]["preferred"]
    # Current or preferred media.
    if album_info.media:
        # Preferred media options.
        media_patterns: Sequence[str] = preferred_config["media"].as_str_seq()
        options = [
            re.compile(rf"(\d+x)?({pat})", re.I) for pat in media_patterns
        ]
        if options:
            dist.add_priority("media", album_info.media, options)
        # Current media.
        elif likelies["media"]:
            dist.add_equality("media", album_info.media, likelies["media"])

    # Mediums.
    if likelies["disctotal"] and album_info.mediums:
        dist.add_number("mediums", likelies["disctotal"], album_info.mediums)

    # Prefer earliest release.
    if album_info.year and preferred_config["original_year"]:
        # Assume 1889 (earliest first gramophone discs) if we don't know the
        # original year.
        original = album_info.original_year or 1889
        diff = abs(album_info.year - original)
        diff_max = abs(datetime.date.today().year - original)
        dist.add_ratio("year", diff, diff_max)
    # Year.
    elif likelies["year"] and album_info.year:
        if likelies["year"] in (album_info.year, album_info.original_year):
            # No penalty for matching release or original year.
            dist.add("year", 0.0)
        elif album_info.original_year:
            # Prefer matchest closest to the release year.
            diff = abs(likelies["year"] - album_info.year)
            diff_max = abs(
                datetime.date.today().year - album_info.original_year
            )
            dist.add_ratio("year", diff, diff_max)
        else:
            # Full penalty when there is no original year.
            dist.add("year", 1.0)

    # Preferred countries.
    country_patterns: Sequence[str] = preferred_config["countries"].as_str_seq()
    options = [re.compile(pat, re.I) for pat in country_patterns]
    if album_info.country and options:
        dist.add_priority("country", album_info.country, options)
    # Country.
    elif likelies["country"] and album_info.country:
        dist.add_string("country", likelies["country"], album_info.country)

    # Label.
    if likelies["label"] and album_info.label:
        dist.add_string("label", likelies["label"], album_info.label)

    # Catalog number.
    if likelies["catalognum"] and album_info.catalognum:
        dist.add_string(
            "catalognum", likelies["catalognum"], album_info.catalognum
        )

    # Disambiguation.
    if likelies["albumdisambig"] and album_info.albumdisambig:
        dist.add_string(
            "albumdisambig", likelies["albumdisambig"], album_info.albumdisambig
        )

    # Album ID.
    if likelies["mb_albumid"]:
        dist.add_equality(
            "album_id", likelies["mb_albumid"], album_info.album_id
        )

    # Tracks.
    dist.tracks = {}
    for item, track in item_info_pairs:
        dist.tracks[track] = track_distance(item, track, album_info.va)
        dist.add("tracks", dist.tracks[track].distance)

    # Missing tracks.
    for _ in range(len(album_info.tracks) - len(item_info_pairs)):
        dist.add("missing_tracks", 1.0)

    # Unmatched tracks.
    for _ in range(len(items) - len(item_info_pairs)):
        dist.add("unmatched_tracks", 1.0)

    dist.add_data_source(likelies["data_source"], album_info.data_source)

    return dist



"""Tests the facility that lets plugins add custom field to MediaFile."""

import shutil

import mediafile
import pytest

from beets.library import Item
from beets.plugins import BeetsPlugin
from beets.test import _common
from beets.test.helper import BeetsTestCase
from beets.util import syspath

field_extension = mediafile.MediaField(
    mediafile.MP3DescStorageStyle("customtag"),
    mediafile.MP4StorageStyle("----:com.apple.iTunes:customtag"),
    mediafile.StorageStyle("customtag"),
    mediafile.ASFStorageStyle("customtag"),
)

list_field_extension = mediafile.ListMediaField(
    mediafile.MP3ListDescStorageStyle("customlisttag"),
    mediafile.MP4ListStorageStyle("----:com.apple.iTunes:customlisttag"),
    mediafile.ListStorageStyle("customlisttag"),
    mediafile.ASFStorageStyle("customlisttag"),
)


class ExtendedFieldTestMixin(BeetsTestCase):
    def _mediafile_fixture(self, name, extension="mp3"):
        name = f"{name}.{extension}"
        src = _common.RSRC / name
        target = self.temp_path / name
        shutil.copy(syspath(src), syspath(target))
        return mediafile.MediaFile(target)

    def test_extended_field_write(self):
        plugin = BeetsPlugin()
        plugin.add_media_field("customtag", field_extension)

        try:
            mf = self._mediafile_fixture("empty")
            mf.customtag = "F#"
            mf.save()

            mf = mediafile.MediaFile(mf.path)
            assert mf.customtag == "F#"

        finally:
            delattr(mediafile.MediaFile, "customtag")
            Item._media_fields.remove("customtag")

    def test_extended_list_field_write(self):
        plugin = BeetsPlugin()
        plugin.add_media_field("customlisttag", list_field_extension)

        try:
            mf = self._mediafile_fixture("empty")
            mf.customlisttag = ["a", "b"]
            mf.save()

            mf = mediafile.MediaFile(mf.path)
            assert mf.customlisttag == ["a", "b"]

        finally:
            delattr(mediafile.MediaFile, "customlisttag")
            Item._media_fields.remove("customlisttag")

    def test_write_extended_tag_from_item(self):
        plugin = BeetsPlugin()
        plugin.add_media_field("customtag", field_extension)

        try:
            mf = self._mediafile_fixture("empty")
            assert mf.customtag is None

            item = Item(path=mf.path, customtag="Gb")
            item.write()
            mf = mediafile.MediaFile(mf.path)
            assert mf.customtag == "Gb"

        finally:
            delattr(mediafile.MediaFile, "customtag")
            Item._media_fields.remove("customtag")

    def test_read_flexible_attribute_from_file(self):
        plugin = BeetsPlugin()
        plugin.add_media_field("customtag", field_extension)

        try:
            mf = self._mediafile_fixture("empty")
            mf.update({"customtag": "F#"})
            mf.save()

            item = Item.from_path(mf.path)
            assert item["customtag"] == "F#"

        finally:
            delattr(mediafile.MediaFile, "customtag")
            Item._media_fields.remove("customtag")

    def test_invalid_descriptor(self):
        with pytest.raises(
            ValueError, match="must be an instance of MediaField"
        ):
            mediafile.MediaFile.add_field("somekey", True)

    def test_overwrite_property(self):
        with pytest.raises(
            ValueError, match='property "artist" already exists'
        ):
            mediafile.MediaFile.add_field("artist", mediafile.MediaField())



from __future__ import annotations

import re
from contextlib import contextmanager
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, TypeVar

import platformdirs

import beets
from beets import config, context, dbcore
from beets.dbcore.query import Query
from beets.dbcore.sort import NullSort
from beets.exceptions import UserError
from beets.util import normpath
from beets.util.pathformats import get_path_formats

from . import migrations
from .models import Album, Item
from .queries import parse_query_parts, parse_query_string

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

    from beets.dbcore.sort import Sort
    from beets.util import PathLike, Replacements
    from beets.util.pathformats import PathFormat

    from .models import LibModel

    LM = TypeVar("LM", bound=LibModel)


class Library(dbcore.Database):
    """A database of music containing songs and albums."""

    _models = (Item, Album)
    _migrations = (
        (migrations.MultiGenreFieldMigration, (Item, Album)),
        (migrations.LyricsMetadataInFlexFieldsMigration, (Item,)),
        (migrations.MultiRemixerFieldMigration, (Item,)),
        (migrations.MultiLyricistFieldMigration, (Item,)),
        (migrations.MultiComposerFieldMigration, (Item,)),
        (migrations.MultiArrangerFieldMigration, (Item,)),
        (migrations.RelativePathMigration, (Item, Album)),
        (migrations.RemoveInheritedArtpathMigration, (Item,)),
        (migrations.InstrumentalLyricsInFlexFieldMigration, (Item,)),
    )

    # Used for template substitution performance.
    _memotable: dict[tuple[str | None, str | None, str | None, int | None], str]
    replacements: Replacements

    @cached_property
    def path_formats(self) -> list[PathFormat]:
        return get_path_formats(config["paths"])

    @staticmethod
    def get_replacements() -> Replacements:
        """Build regex/string replacement pairs from config."""
        replacements = []
        for pattern, repl in beets.config["replace"].get(dict).items():
            repl = repl or ""
            try:
                replacements.append((re.compile(pattern), repl))
            except re.error:
                raise UserError(
                    f"Malformed regular expression in replace: {pattern}"
                )
        return replacements

    def __init__(
        self,
        path: PathLike = Path("library.blb"),
        directory: str | None = None,
        set_music_dir: bool = True,
    ) -> None:
        self.directory = normpath(directory or platformdirs.user_music_path())
        if set_music_dir:
            context.set_music_dir(self.directory)

        super().__init__(path, timeout=beets.config["timeout"].as_number())

        self.replacements = self.get_replacements()
        self._memotable = {}

    @contextmanager
    def music_dir_context(self) -> Iterator[Library]:
        """Temporarily bind this library's directory to path conversion."""
        with context.music_dir(self.directory):
            yield self

    # Adding objects to the database.

    def add(self, obj: LibModel) -> int | None:
        """Add the :class:`Item` or :class:`Album` object to the library
        database.

        Return the object's new id.
        """
        obj.add(self)
        self._memotable = {}
        return obj.id

    def add_album(self, items: list[Item]) -> Album:
        """Create a new album consisting of a list of items.

        The items are added to the database if they don't yet have an
        ID. Return a new :class:`Album` object. The list items must not
        be empty.
        """
        if not items:
            raise ValueError("need at least one item")

        # Create the album structure using metadata from the first item.
        values = {key: items[0][key] for key in Album.item_keys}
        album = Album(self, **values)

        # Add the album structure and set the items' album_id fields.
        # Store or add the items.
        with self.transaction():
            album.add(self)
            for item in items:
                item.album_id = album.id
                if item.id is None:
                    item.add(self)
                else:
                    item.store()

        return album

    # Querying.

    def _fetch(
        self,
        model_cls: type[LM],
        query: str | Sequence[str] | Query | None = None,
        sort: Sort | None = None,
    ) -> dbcore.Results[LM]:
        """Parse a query and fetch.

        If an order specification is present in the query string
        the `sort` argument is ignored.
        """
        # Parse the query, if necessary.
        parsed_sort = None
        parsed_query = None
        try:
            # Query parsing needs the library root, but keeping it scoped here
            # avoids leaking one Library's directory into another's work.
            with context.music_dir(self.directory):
                if isinstance(query, Query):
                    parsed_query = query
                if isinstance(query, str):
                    parsed_query, parsed_sort = parse_query_string(
                        query, model_cls
                    )
                elif isinstance(query, (list, tuple)):
                    parsed_query, parsed_sort = parse_query_parts(
                        query, model_cls
                    )
        except dbcore.query.InvalidQueryArgumentValueError as exc:
            raise dbcore.InvalidQueryError(query, exc)

        # Any non-null sort specified by the parsed query overrides the
        # provided sort.
        if parsed_sort and not isinstance(parsed_sort, NullSort):
            sort = parsed_sort

        return super()._get_results(model_cls, parsed_query, sort)

    @staticmethod
    def get_default_album_sort() -> Sort:
        """Get a :class:`Sort` object for albums from the config option."""
        return dbcore.sort_from_strings(
            Album, beets.config["sort_album"].as_str_seq()
        )

    @staticmethod
    def get_default_item_sort() -> Sort:
        """Get a :class:`Sort` object for items from the config option."""
        return dbcore.sort_from_strings(
            Item, beets.config["sort_item"].as_str_seq()
        )

    def albums(
        self,
        query: str | Sequence[str] | Query | None = None,
        sort: Sort | None = None,
    ) -> dbcore.Results[Album]:
        """Get :class:`Album` objects matching the query."""
        return self._fetch(Album, query, sort or self.get_default_album_sort())

    def items(
        self,
        query: str | Sequence[str] | Query | None = None,
        sort: Sort | None = None,
    ) -> dbcore.Results[Item]:
        """Get :class:`Item` objects matching the query."""
        return self._fetch(Item, query, sort or self.get_default_item_sort())

    # Convenience accessors.
    def get_item(self, id_: int) -> Item | None:
        """Fetch a :class:`Item` by its ID.

        Return `None` if no match is found.
        """
        return self._get(Item, id_)

    def get_album(self, item_or_id: Item | int) -> Album | None:
        """Given an album ID or an item associated with an album, return
        a :class:`Album` object for the album.

        If no such album exists, return `None`.
        """
        album_id = (
            item_or_id if isinstance(item_or_id, int) else item_or_id.album_id
        )
        return self._get(Album, album_id) if album_id else None
